# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from datetime import date
from typing import Dict, List, Set, cast
from rp2.abstract_entry import AbstractEntry

from rp2.balance import BalanceSet
from rp2.configuration import MAX_DATE, MIN_DATE, Configuration
from rp2.entry_types import EntrySetType, TransactionType
from rp2.gain_loss import GainLoss
from rp2.gain_loss_set import GainLossSet
from rp2.in_transaction import InTransaction
from rp2.input_data import InputData
from rp2.intra_transaction import IntraTransaction
from rp2.logger import LOGGER
from rp2.out_transaction import OutTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError
from rp2.transaction_set import TransactionSet


@dataclass(frozen=True, eq=True)
class YearlyGainLoss:
    year: int
    asset: str
    transaction_type: TransactionType
    is_long_term_capital_gains: bool
    crypto_amount: RP2Decimal
    fiat_amount: RP2Decimal
    fiat_cost_basis: RP2Decimal
    fiat_gain_loss: RP2Decimal

    def __post_init__(self) -> None:
        Configuration.type_check_positive_int("year", self.year)
        Configuration.type_check_string("asset", self.asset)
        TransactionType.type_check("transaction_type", self.transaction_type)
        Configuration.type_check_bool("is_long_term", self.is_long_term_capital_gains)
        Configuration.type_check_decimal("crypto_amount", self.crypto_amount)
        Configuration.type_check_decimal("fiat_amount", self.fiat_amount)
        Configuration.type_check_decimal("fiat_cost_basis", self.fiat_cost_basis)
        Configuration.type_check_decimal("fiat_gain_loss", self.fiat_gain_loss)

    def __eq__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, YearlyGainLoss):
            raise RP2TypeError(f"Operand has non-YearlyGainLoss value {repr(other)}")
        result: bool = (
            self.year == other.year
            and self.asset == other.asset
            and self.transaction_type == other.transaction_type
            and self.is_long_term_capital_gains == other.is_long_term_capital_gains
        )
        return result

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __lt__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, YearlyGainLoss):
            raise RP2TypeError(f"Operand has non-YearlyGainLoss value {repr(other)}")
        return self.year < other.year

    def __gt__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, YearlyGainLoss):
            raise RP2TypeError(f"Operand has non-YearlyGainLoss value {repr(other)}")
        return self.year > other.year

    def __ge__(self, other: object) -> bool:
        return not self.__lt__(other)

    def __le__(self, other: object) -> bool:
        return not self.__gt__(other)

    def __hash__(self) -> int:
        return hash((self.year, self.asset, self.transaction_type, self.is_long_term_capital_gains))


@dataclass(frozen=True, eq=True)
class _YearlyGainLossId:
    year: int
    asset: str
    transaction_type: TransactionType
    is_long_term_capital_gains: bool


# Frozen and eq are not set because we don't need to hash instances and we need to modify fields (see _create_yearly_gain_loss_list)
@dataclass(frozen=True, eq=True)
class _YearlyGainLossAmounts:
    crypto_amount: RP2Decimal
    fiat_amount: RP2Decimal
    fiat_cost_basis: RP2Decimal
    fiat_gain_loss: RP2Decimal


class ComputedData:
    @classmethod
    def type_check(cls, name: str, instance: "ComputedData") -> "ComputedData":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    @staticmethod
    def _compute_price_per_unit(unfiltered_in_transaction_set: TransactionSet, to_date: date) -> RP2Decimal:
        crypto_in_running_sum: RP2Decimal = ZERO
        fiat_in_with_fee_running_sum: RP2Decimal = ZERO
        for entry in unfiltered_in_transaction_set:
            # from_date is not used when computing average price per unit (because we always start from the beginning): only to_date is relevant.
            if entry.timestamp.date() > to_date:
                break
            transaction: InTransaction = cast(InTransaction, entry)
            crypto_in_running_sum += transaction.crypto_in
            fiat_in_with_fee_running_sum += transaction.fiat_in_with_fee
        return fiat_in_with_fee_running_sum / crypto_in_running_sum if crypto_in_running_sum is not ZERO else ZERO

    @staticmethod
    def _filter_yearly_gain_loss_by_year(unfiltered_yearly_gain_loss_list: List[YearlyGainLoss], from_year: int) -> List[YearlyGainLoss]:
        return [y for y in unfiltered_yearly_gain_loss_list if y.year >= from_year]

    # from_date is not used when computing yearly gain loss (because we always start at the beginning): only to_date is relevant.
    @staticmethod
    def _create_yearly_gain_loss_list(
        unfiltered_gain_loss_set: GainLossSet,
        to_date: date = MAX_DATE,
    ) -> List[YearlyGainLoss]:
        summaries: Dict[_YearlyGainLossId, _YearlyGainLossAmounts] = {}
        entry: AbstractEntry
        key: _YearlyGainLossId
        value: _YearlyGainLossAmounts
        for entry in unfiltered_gain_loss_set:
            gain_loss: GainLoss = cast(GainLoss, entry)
            if gain_loss.taxable_event.timestamp.date() > to_date:
                break
            key = _YearlyGainLossId(
                gain_loss.taxable_event.timestamp.year,
                gain_loss.asset,
                gain_loss.taxable_event.transaction_type,
                gain_loss.is_long_term_capital_gains(),
            )
            value = summaries.setdefault(key, _YearlyGainLossAmounts(ZERO, ZERO, ZERO, ZERO))
            crypto_amount: RP2Decimal = value.crypto_amount + gain_loss.crypto_amount
            fiat_amount: RP2Decimal = value.fiat_amount + gain_loss.taxable_event_fiat_amount_with_fee_fraction
            fiat_cost_basis: RP2Decimal = value.fiat_cost_basis + gain_loss.fiat_cost_basis
            fiat_gain_loss: RP2Decimal = value.fiat_gain_loss + gain_loss.fiat_gain
            summaries[key] = _YearlyGainLossAmounts(
                crypto_amount=crypto_amount, fiat_amount=fiat_amount, fiat_cost_basis=fiat_cost_basis, fiat_gain_loss=fiat_gain_loss
            )

        yearly_gain_loss_set: Set[YearlyGainLoss] = set()
        crypto_taxable_amount_total: RP2Decimal = ZERO
        fiat_taxable_amount_total: RP2Decimal = ZERO
        cost_basis_total: RP2Decimal = ZERO
        gain_loss_total: RP2Decimal = ZERO
        for (key, value) in summaries.items():
            yearly_gain_loss: YearlyGainLoss = YearlyGainLoss(
                year=key.year,
                asset=key.asset,
                transaction_type=key.transaction_type,
                is_long_term_capital_gains=key.is_long_term_capital_gains,
                crypto_amount=value.crypto_amount,
                fiat_amount=value.fiat_amount,
                fiat_cost_basis=value.fiat_cost_basis,
                fiat_gain_loss=value.fiat_gain_loss,
            )
            yearly_gain_loss_set.add(yearly_gain_loss)
            crypto_taxable_amount_total += yearly_gain_loss.crypto_amount
            fiat_taxable_amount_total += yearly_gain_loss.fiat_amount
            cost_basis_total += yearly_gain_loss.fiat_cost_basis
            gain_loss_total += yearly_gain_loss.fiat_gain_loss

        return list(sorted(yearly_gain_loss_set, key=_yearly_gain_loss_sort_criteria, reverse=True))

    def __init__(
        self,
        asset: str,
        unfiltered_taxable_event_set: TransactionSet,
        unfiltered_gain_loss_set: GainLossSet,
        input_data: InputData,
        from_date: date = MIN_DATE,
        to_date: date = MAX_DATE,
    ) -> None:
        # pylint: disable=too-many-branches
        InputData.type_check("input_data", input_data)
        if not isinstance(from_date, date):
            raise RP2TypeError("Parameter 'from_date' is not of type date")
        if not isinstance(to_date, date):
            raise RP2TypeError("Parameter 'to_date' is not of type date")

        self.__asset: str = Configuration.type_check_string("asset", asset)
        TransactionSet.type_check("taxable_event_set", unfiltered_taxable_event_set, EntrySetType.MIXED, asset, True)
        GainLossSet.type_check("gain_loss_set", unfiltered_gain_loss_set)

        self.__filtered_taxable_event_set: TransactionSet = cast(TransactionSet, unfiltered_taxable_event_set.duplicate(from_date=from_date, to_date=to_date))
        self.__filtered_gain_loss_set: GainLossSet = cast(GainLossSet, unfiltered_gain_loss_set.duplicate(from_date=from_date, to_date=to_date))

        yearly_gain_loss_list: List[YearlyGainLoss] = self._create_yearly_gain_loss_list(unfiltered_gain_loss_set, to_date)
        LOGGER.debug("%s: Created yearly gain-loss list", input_data.asset)
        self.__filtered_yearly_gain_loss_list: List[YearlyGainLoss] = self._filter_yearly_gain_loss_by_year(yearly_gain_loss_list, from_date.year)

        self.__filtered_in_transaction_set: TransactionSet = input_data.filtered_in_transaction_set
        self.__filtered_intra_transaction_set: TransactionSet = input_data.filtered_intra_transaction_set
        self.__filtered_out_transaction_set: TransactionSet = input_data.filtered_out_transaction_set

        self.__filtered_balance_set: BalanceSet = BalanceSet(unfiltered_taxable_event_set.configuration, input_data, to_date)
        self.__filtered_price_per_unit: RP2Decimal = self._compute_price_per_unit(input_data.unfiltered_in_transaction_set, to_date)

        # Compute crypto running sums
        self.__crypto_in_running_sum: Dict[InTransaction, RP2Decimal] = {}
        self.__crypto_out_running_sum: Dict[OutTransaction, RP2Decimal] = {}
        self.__crypto_out_fee_running_sum: Dict[OutTransaction, RP2Decimal] = {}
        self.__crypto_intra_fee_running_sum: Dict[IntraTransaction, RP2Decimal] = {}
        self.__crypto_gain_loss_running_sum: Dict[GainLoss, RP2Decimal] = {}

        crypto_running_sum: RP2Decimal
        crypto_fee_running_sum: RP2Decimal

        crypto_running_sum = ZERO
        for entry in input_data.unfiltered_in_transaction_set:
            in_transaction: InTransaction = cast(InTransaction, entry)
            crypto_running_sum += in_transaction.crypto_in
            self.__crypto_in_running_sum[in_transaction] = crypto_running_sum

        crypto_running_sum = ZERO
        crypto_fee_running_sum = ZERO
        for entry in input_data.unfiltered_out_transaction_set:
            out_transaction: OutTransaction = cast(OutTransaction, entry)
            crypto_running_sum += out_transaction.crypto_out_no_fee
            crypto_fee_running_sum += out_transaction.crypto_fee
            self.__crypto_out_running_sum[out_transaction] = crypto_running_sum
            self.__crypto_out_fee_running_sum[out_transaction] = crypto_fee_running_sum

        crypto_fee_running_sum = ZERO
        for entry in input_data.unfiltered_intra_transaction_set:
            intra_transaction: IntraTransaction = cast(IntraTransaction, entry)
            crypto_fee_running_sum += intra_transaction.crypto_fee
            self.__crypto_intra_fee_running_sum[intra_transaction] = crypto_fee_running_sum

        crypto_running_sum = ZERO
        gain_loss: GainLoss
        for entry in unfiltered_gain_loss_set:
            gain_loss = cast(GainLoss, entry)
            crypto_running_sum += gain_loss.crypto_amount
            self.__crypto_gain_loss_running_sum[gain_loss] = crypto_running_sum

        # Compute in lot sold percentages
        self.__in_lot_sold_percentage: Dict[InTransaction, RP2Decimal] = {}
        for entry in self.__filtered_gain_loss_set:
            gain_loss = cast(GainLoss, entry)
            if not gain_loss.acquired_lot or gain_loss.acquired_lot.timestamp.date() < from_date or gain_loss.acquired_lot.timestamp.date() > to_date:
                continue
            self.__in_lot_sold_percentage[gain_loss.acquired_lot] = (
                self.__in_lot_sold_percentage.setdefault(gain_loss.acquired_lot, ZERO) + gain_loss.acquired_lot_fraction_percentage
            )

        if self.__filtered_taxable_event_set.asset != self.__asset:
            raise RP2ValueError(f"Asset mismatch in 'taxable_event_set': expected {self.__asset}, found {self.__filtered_taxable_event_set.asset}")
        if self.__filtered_gain_loss_set.asset != self.__asset:
            raise RP2ValueError(f"Asset mismatch in 'gain_loss_set': expected {self.__asset}, found {self.__filtered_gain_loss_set.asset}")
        if self.__filtered_balance_set.asset != self.__asset:
            raise RP2ValueError(f"Asset mismatch in 'balance_set': expected {self.__asset}, found {self.__filtered_balance_set.asset}")

        if self.__asset != input_data.asset:
            raise RP2ValueError(f"Asset mismatch in 'input_data': expected {self.__asset}, found {input_data.asset}")

    @property
    def asset(self) -> str:
        """Asset this ComputedData instance is about."""
        return self.__asset

    @property
    def taxable_event_set(self) -> TransactionSet:
        """Set of taxable events in this ComputedData instance."""
        return self.__filtered_taxable_event_set

    @property
    def gain_loss_set(self) -> GainLossSet:
        """Set of gain/loss mappings in this ComputedData instance."""
        return self.__filtered_gain_loss_set

    @property
    def yearly_gain_loss_list(self) -> List[YearlyGainLoss]:
        """List of gain/loss summaries in this ComputedData instance, grouped by year."""
        return self.__filtered_yearly_gain_loss_list

    @property
    def in_transaction_set(self) -> TransactionSet:
        """Set of in-transactions in this ComputedData instance."""
        return self.__filtered_in_transaction_set

    @property
    def out_transaction_set(self) -> TransactionSet:
        """Set of out-transactions in this ComputedData instance."""
        return self.__filtered_out_transaction_set

    @property
    def intra_transaction_set(self) -> TransactionSet:
        """Set of intra-transactions in this ComputedData instance."""
        return self.__filtered_intra_transaction_set

    @property
    def balance_set(self) -> BalanceSet:
        """Set of account balances in this ComputedData instance."""
        return self.__filtered_balance_set

    @property
    def price_per_unit(self) -> RP2Decimal:
        """Average price per asset unit."""
        return self.__filtered_price_per_unit

    def get_crypto_in_running_sum(self, in_transaction: InTransaction) -> RP2Decimal:
        """Crypto in running sum for a given InTransaction instance."""
        InTransaction.type_check("in_transaction", in_transaction)
        return self.__crypto_in_running_sum[in_transaction]

    def get_crypto_out_running_sum(self, out_transaction: OutTransaction) -> RP2Decimal:
        """Crypto running sum for a given OutTransaction instance."""
        OutTransaction.type_check("out_transaction", out_transaction)
        return self.__crypto_out_running_sum[out_transaction]

    def get_crypto_out_fee_running_sum(self, out_transaction: OutTransaction) -> RP2Decimal:
        """Crypto fee running sum for a given OutTransaction instance."""
        OutTransaction.type_check("out_transaction", out_transaction)
        return self.__crypto_out_fee_running_sum[out_transaction]

    def get_crypto_intra_fee_running_sum(self, intra_transaction: IntraTransaction) -> RP2Decimal:
        """Crypto fee running sum for a given IntraTransaction instance."""
        IntraTransaction.type_check("intra_transaction", intra_transaction)
        return self.__crypto_intra_fee_running_sum[intra_transaction]

    def get_crypto_gain_loss_running_sum(self, gain_loss: GainLoss) -> RP2Decimal:
        """Crypto amount running sum for a given GainLoss instance."""
        GainLoss.type_check("gain_loss", gain_loss)
        return self.__crypto_gain_loss_running_sum[gain_loss]

    def get_in_lot_sold_percentage(self, in_transaction: InTransaction) -> RP2Decimal:
        """Percentage sold for a given InTransaction instance"""
        InTransaction.type_check("in_transaction", in_transaction)
        return self.__in_lot_sold_percentage[in_transaction] if in_transaction in self.__in_lot_sold_percentage else ZERO


def _yearly_gain_loss_sort_criteria(yearly_gain_loss: YearlyGainLoss) -> str:
    return (
        f"{yearly_gain_loss.asset}"
        f" {yearly_gain_loss.year}"
        f" {'LONG' if yearly_gain_loss.is_long_term_capital_gains else 'SHORT'}"
        f" {yearly_gain_loss.transaction_type.value}"
    )



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from datetime import date, datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

from dateutil.parser import parse
from jsonschema import validate
from rp2.abstract_country import AbstractCountry
from rp2.configuration_schema import CONFIGURATION_SCHEMA
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError

VERSION: str = "0.9.12"

MIN_DATE: date = date(1970, 1, 1)
MAX_DATE: date = date(9999, 12, 31)

# Parametrized and extensible method to generate string representation
def to_string(indent: int = 0, repr_format: bool = True, data: Optional[List[str]] = None) -> str:
    padding: str
    output: List[str] = []
    separator: str
    if not data:
        return ""

    if repr_format:
        padding = ""
        separator = ", "
        data[0] = f"{'  ' * indent}{data[0]}"
    else:
        padding = "  " * indent
        separator = "\n  "

    if data:
        for line in data:
            output.append(f"{padding}{line}")

    if repr_format:
        output[-1] += ")"

    # Joining by separator adds one level of indentation to internal fields (like id) in str mode, which is correct.
    return separator.join(output)


class Configuration:  # pylint: disable=too-many-public-methods
    @classmethod
    def type_check(cls, name: str, instance: "Configuration") -> "Configuration":
        cls.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __init__(
        self,
        configuration_path: str,
        country: AbstractCountry,
        from_date: date = MIN_DATE,
        to_date: date = MAX_DATE,
    ) -> None:
        self.__configuration_path: str = self.type_check_string("configuration_path", configuration_path)
        self.__country = AbstractCountry.type_check("country", country)
        if not isinstance(from_date, date):
            raise RP2TypeError("Parameter 'from_date' is not of type date")
        self.__from_date: date = from_date
        if not isinstance(to_date, date):
            raise RP2TypeError("Parameter 'to_date' is not of type date")
        self.__to_date: date = to_date

        if self.__from_date > self.__to_date:
            raise RP2ValueError("Parameter from_date cannot be greater than to_date")

        self.__in_header: Dict[str, int]
        self.__out_header: Dict[str, int]
        self.__intra_header: Dict[str, int]
        self.__assets: Set[str]
        self.__exchanges: Set[str]
        self.__holders: Set[str]

        if not Path(configuration_path).exists():
            raise RP2ValueError(f"Error: {configuration_path} does not exist")

        with open(configuration_path, encoding="utf-8") as configuration_file:
            # This json_configuration is validated by jsonschema, so we can disable static type checking for it:
            # it adds complexity but not much value over jsonschema checks
            json_configuration: Any = json.load(configuration_file)
            validate(instance=json_configuration, schema=CONFIGURATION_SCHEMA)

            self.__in_header = json_configuration["in_header"]
            self.__out_header = json_configuration["out_header"]
            self.__intra_header = json_configuration["intra_header"]
            self.__assets = set(json_configuration["assets"])
            self.__exchanges = set(json_configuration["exchanges"])
            self.__holders = set(json_configuration["holders"])

        # Used by __repr__()
        self.__sorted_assets: List[str] = sorted(self.__assets)
        self.__sorted_exchanges: List[str] = sorted(self.__exchanges)
        self.__sorted_holders: List[str] = sorted(self.__holders)

    def __repr__(self) -> str:
        return (
            f"Configuration(configuration_path={self.configuration_path}, "
            f"country={repr(self.country)}, "
            f"from_date={self.from_date if self.from_date > MIN_DATE else 'non-specified'}, "
            f"to_date={self.to_date if self.to_date < MAX_DATE else 'non-specified'}, "
            f"in_header={self.__in_header}, "
            f"out_header={self.__out_header}, "
            f"intra_header={self.__intra_header}, "
            f"assets={self.__sorted_assets}, "
            f"exchanges={self.__sorted_exchanges}, "
            f"holders={self.__sorted_holders})"
        )

    @property
    def configuration_path(self) -> str:
        return self.__configuration_path

    @property
    def country(self) -> AbstractCountry:
        return self.__country

    @property
    def from_date(self) -> date:
        return self.__from_date

    @property
    def to_date(self) -> date:
        return self.__to_date

    @property
    def assets(self) -> Set[str]:
        return self.__assets

    def __get_table_constructor_argument_pack(self, data: List[Any], table_type: str, header: Dict[str, int]) -> Dict[str, Any]:
        if not isinstance(data, List):
            raise RP2TypeError(f"Parameter 'data' value is not a List: {data}")
        max_column: int = header[max(header, key=header.get)]  # type: ignore
        if len(data) <= max_column:
            raise RP2ValueError(
                f"Parameter 'data' has length {len(data)}, but required minimum from {table_type}-table headers in "
                f"{self.__configuration_path} is {max_column + 1}"
            )
        pack: Dict[str, Any] = {argument: data[position] for argument, position in header.items()}

        return pack

    def get_in_table_constructor_argument_pack(self, data: List[Any]) -> Dict[str, Any]:
        return self.__get_table_constructor_argument_pack(data, "in", self.__in_header)

    def get_out_table_constructor_argument_pack(self, data: List[Any]) -> Dict[str, Any]:
        return self.__get_table_constructor_argument_pack(data, "out", self.__out_header)

    def get_intra_table_constructor_argument_pack(self, data: List[Any]) -> Dict[str, Any]:
        return self.__get_table_constructor_argument_pack(data, "intra", self.__intra_header)

    def get_in_table_column_position(self, input_parameter: str) -> int:
        self.type_check_string("input_parameter", input_parameter)
        if input_parameter not in self.__in_header:
            raise RP2ValueError(f"Unknown 'input_parameter' {input_parameter}")
        return self.__in_header[input_parameter]

    def get_out_table_column_position(self, input_parameter: str) -> int:
        self.type_check_string("input_parameter", input_parameter)
        if input_parameter not in self.__out_header:
            raise RP2ValueError(f"Unknown 'input_parameter' {input_parameter}")
        return self.__out_header[input_parameter]

    def get_intra_table_column_position(self, input_parameter: str) -> int:
        self.type_check_string("input_parameter", input_parameter)
        if input_parameter not in self.__intra_header:
            raise RP2ValueError(f"Unknown 'input_parameter' {input_parameter}")
        return self.__intra_header[input_parameter]

    @classmethod
    def type_check_unique_id(cls, name: str, value: int) -> int:
        return cls.type_check_positive_int(name, value)

    @classmethod
    def type_check_timestamp_from_string(cls, name: str, value: str) -> datetime:
        cls.type_check_string(name, value)
        try:
            result: datetime = parse(value)
        except Exception as exc:
            raise RP2ValueError(f"Error parsing parameter '{name}': {str(exc)}") from exc
        if result.tzinfo is None:
            raise RP2ValueError(f"Parameter '{name}' value has no timezone info: {value}")
        return result

    def type_check_exchange(self, name: str, value: str) -> str:
        self.type_check_string(name, value)
        if value not in self.__exchanges:
            raise RP2ValueError(f"Parameter '{name}' value is not known: {value}")
        return value

    def type_check_holder(self, name: str, value: str) -> str:
        self.type_check_string(name, value)
        if value not in self.__holders:
            raise RP2ValueError(f"Parameter '{name}' value is not known: {value}")
        return value

    def type_check_asset(self, name: str, value: str) -> str:
        self.type_check_string(name, value)
        if value not in self.__assets:
            raise RP2ValueError(f"Parameter '{name}' value is not known: {value}")
        return value

    @classmethod
    def type_check_parameter_name(cls, name: str) -> str:
        if not isinstance(name, str):
            raise RP2TypeError(f"Parameter name is not a string: {name}")
        return name

    @classmethod
    def type_check_string(cls, name: str, value: str) -> str:
        cls.type_check_parameter_name(name)
        if not isinstance(value, str):
            raise RP2TypeError(f"Parameter '{name}' has non-string value {repr(value)}")
        return value

    @classmethod
    def type_check_positive_int(cls, name: str, value: int, non_zero: bool = False) -> int:
        result: int = cls.type_check_int(name, value)
        if result < 0:
            raise RP2ValueError(f"Parameter '{name}' has non-positive value {value}")
        if non_zero and result == 0:
            raise RP2ValueError(f"Parameter '{name}' has zero value")
        return result

    @classmethod
    def type_check_int(cls, name: str, value: int) -> int:
        cls.type_check_parameter_name(name)
        if not isinstance(value, int):
            raise RP2TypeError(f"Parameter '{name}' has non-integer value {repr(value)}")
        return value

    @classmethod
    def type_check_positive_float(cls, name: str, value: float, non_zero: bool = False) -> float:
        result: float = cls.type_check_float(name, value)
        if result < 0:
            raise RP2ValueError(f"Parameter '{name}' has non-positive value {value}")
        if non_zero and result == 0:
            raise RP2ValueError(f"Parameter '{name}' has zero value")
        return result

    @classmethod
    def type_check_float(cls, name: str, value: float) -> float:
        cls.type_check_parameter_name(name)
        if not isinstance(value, (int, float)):
            raise RP2TypeError(f"Parameter '{name}' has non-numeric value {repr(value)}")
        return value

    @classmethod
    def type_check_bool(cls, name: str, value: bool) -> bool:
        cls.type_check_parameter_name(name)
        if not isinstance(value, bool):
            raise RP2TypeError(f"Parameter '{name}' has non-bool value {repr(value)}")
        return value

    @classmethod
    def type_check_positive_decimal(cls, name: str, value: RP2Decimal, non_zero: bool = False) -> RP2Decimal:
        result: RP2Decimal = cls.type_check_decimal(name, value)
        if result < ZERO:
            raise RP2ValueError(f"Parameter '{name}' has non-positive value {value}")
        if non_zero and result == ZERO:
            raise RP2ValueError(f"Parameter '{name}' has zero value")
        return result

    @classmethod
    def type_check_decimal(cls, name: str, value: RP2Decimal) -> RP2Decimal:
        cls.type_check_parameter_name(name)
        if not isinstance(value, RP2Decimal):
            raise RP2TypeError(f"Parameter '{name}' has non-RP2Decimal value {repr(value)}")
        return value



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import date
from typing import Dict

from rp2.abstract_country import AbstractCountry
from rp2.computed_data import ComputedData


class AbstractReportGenerator:
    def generate(
        self,
        country: AbstractCountry,
        accounting_method: str,
        asset_to_computed_data: Dict[str, ComputedData],
        output_dir_path: str,
        output_file_prefix: str,
        from_date: date,
        to_date: date,
    ) -> None:
        raise NotImplementedError("Abstract method: it must be implemented in the plugin class")

    @classmethod
    def get_name(cls) -> str:
        return f"{cls.__module__.rsplit('.', 1)[1]}"



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterator, NamedTuple, Optional

from rp2.abstract_transaction import AbstractTransaction
from rp2.in_transaction import InTransaction
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError


class TaxableEventAndAcquiredLot(NamedTuple):
    taxable_event: AbstractTransaction
    acquired_lot: Optional[InTransaction]
    taxable_event_amount: RP2Decimal
    acquired_lot_amount: RP2Decimal


class TaxableEventsExhaustedException(Exception):
    def __init__(self, message: str = "") -> None:
        self.__message = message
        super().__init__(self.__message)

    def __repr__(self) -> str:
        return self.message

    @property
    def message(self) -> str:
        return self.__message


class AcquiredLotsExhaustedException(Exception):
    def __init__(self, message: str = "") -> None:
        self.__message = message
        super().__init__(self.__message)

    def __repr__(self) -> str:
        return self.message

    @property
    def message(self) -> str:
        return self.__message


class AbstractAccountingMethod:
    @classmethod
    def type_check(cls, name: str, instance: "AbstractAccountingMethod") -> "AbstractAccountingMethod":
        if not isinstance(name, str):
            raise RP2TypeError(f"Parameter name is not a string: {repr(name)}")
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    # Iterators yield transactions in ascending chronological order
    def initialize(self, taxable_event_iterator: Iterator[AbstractTransaction], acquired_lot_iterator: Iterator[InTransaction]) -> None:
        raise NotImplementedError("Abstract function")

    def get_next_taxable_event_and_amount(
        self,
        taxable_event: Optional[AbstractTransaction],
        acquired_lot: Optional[InTransaction],
        taxable_event_amount: RP2Decimal,
        acquired_lot_amount: RP2Decimal,
    ) -> TaxableEventAndAcquiredLot:
        raise NotImplementedError("Abstract function")

    def get_acquired_lot_for_taxable_event(
        self, taxable_event: AbstractTransaction, acquired_lot: Optional[InTransaction], taxable_event_amount: RP2Decimal, acquired_lot_amount: RP2Decimal
    ) -> TaxableEventAndAcquiredLot:
        raise NotImplementedError("Abstract function")

    def validate_acquired_lot_ancestor_timestamp(self, acquired_lot: InTransaction, acquired_lot_parent: InTransaction) -> bool:
        raise NotImplementedError("Abstract function")

    @property
    def name(self) -> str:
        return f"{self.__class__.__module__.rsplit('.', 1)[1]}"

    def __repr__(self) -> str:
        return self.name



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterator, Optional

from rp2.abstract_accounting_method import (
    AbstractAccountingMethod,
    AcquiredLotsExhaustedException,
    TaxableEventAndAcquiredLot,
    TaxableEventsExhaustedException,
)
from rp2.abstract_transaction import AbstractTransaction
from rp2.in_transaction import InTransaction
from rp2.rp2_decimal import RP2Decimal


# FIFO accounting method. See https://www.investopedia.com/terms/f/fifo.asp.
class AccountingMethod(AbstractAccountingMethod):

    __taxable_event_iterator: Iterator[AbstractTransaction]
    __acquired_lot_iterator: Iterator[InTransaction]

    # Iterators yield transactions in ascending chronological order
    def initialize(self, taxable_event_iterator: Iterator[AbstractTransaction], acquired_lot_iterator: Iterator[InTransaction]) -> None:
        self.__taxable_event_iterator = taxable_event_iterator
        self.__acquired_lot_iterator = acquired_lot_iterator

    def get_next_taxable_event_and_amount(
        self,
        taxable_event: Optional[AbstractTransaction],
        acquired_lot: Optional[InTransaction],
        taxable_event_amount: RP2Decimal,
        acquired_lot_amount: RP2Decimal,
    ) -> TaxableEventAndAcquiredLot:
        try:
            new_taxable_event: AbstractTransaction = next(self.__taxable_event_iterator)
        except StopIteration:
            raise TaxableEventsExhaustedException() from None
        return TaxableEventAndAcquiredLot(
            taxable_event=new_taxable_event,
            acquired_lot=acquired_lot,
            taxable_event_amount=new_taxable_event.crypto_balance_change,
            acquired_lot_amount=acquired_lot_amount - taxable_event_amount,
        )

    def get_acquired_lot_for_taxable_event(
        self, taxable_event: AbstractTransaction, acquired_lot: Optional[InTransaction], taxable_event_amount: RP2Decimal, acquired_lot_amount: RP2Decimal
    ) -> TaxableEventAndAcquiredLot:
        try:
            new_acquired_lot: InTransaction = next(self.__acquired_lot_iterator)
        except StopIteration:
            raise AcquiredLotsExhaustedException() from None
        return TaxableEventAndAcquiredLot(
            taxable_event=taxable_event,
            acquired_lot=new_acquired_lot,
            taxable_event_amount=taxable_event_amount - acquired_lot_amount,
            acquired_lot_amount=new_acquired_lot.crypto_in,
        )

    def validate_acquired_lot_ancestor_timestamp(self, acquired_lot: InTransaction, acquired_lot_parent: InTransaction) -> bool:
        return acquired_lot.timestamp >= acquired_lot_parent.timestamp



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Iterator, List, NamedTuple, Optional

from rp2.abstract_accounting_method import (
    AcquiredLotsExhaustedException,
    TaxableEventAndAcquiredLot,
    TaxableEventsExhaustedException,
)
from rp2.abstract_specific_id import AbstractSpecificId
from rp2.abstract_transaction import AbstractTransaction
from rp2.avl_tree import AVLTree
from rp2.in_transaction import InTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal


@dataclass
class AcquiredLotAndIndex:
    acquired_lot: InTransaction
    index: int


class AcquiredLotAndAmount(NamedTuple):
    acquired_lot: InTransaction
    amount: RP2Decimal


# LIFO plugin. See https://www.investopedia.com/terms/l/lifo.asp. Note that under LIFO the date acquired must still be before or on the date sold:
# see this discussion for details,
# https://ttlc.intuit.com/community/investments-and-rental-properties/discussion/using-lifo-method-for-cryptocurrency-or-even-stock-cost-basis/00/1433542
class AccountingMethod(AbstractSpecificId):

    __taxable_event_iterator: Iterator[AbstractTransaction]
    __year_2_acquired_lot_list: Dict[int, List[InTransaction]]
    __year_2_acquired_lot_avl: Dict[int, AVLTree[str, AcquiredLotAndIndex]]
    __acquired_lot_2_partial_amount: Dict[InTransaction, RP2Decimal]
    __min_acquired_lot_year: int

    # Disambiguation is needed for transactions that have the same timestamp, because the avl tree class expects unique keys: 12 decimal digits express
    # 1 quadrillion, which should be enough to capture the maximum number of same-timestamp transactions in all reasonable cases.
    KEY_DISAMBIGUATOR_LENGTH: int = 12
    MAX_KEY_DISAMBIGUATOR = "9" * KEY_DISAMBIGUATOR_LENGTH

    # Iterators yield transactions in ascending chronological order
    def initialize(self, taxable_event_iterator: Iterator[AbstractTransaction], acquired_lot_iterator: Iterator[InTransaction]) -> None:
        self.__taxable_event_iterator = taxable_event_iterator
        self.__year_2_acquired_lot_list = {}
        self.__year_2_acquired_lot_avl = {}
        self.__acquired_lot_2_partial_amount = {}

        # Initialize data structure to hold acquired_lots in chronological order by year (dictionary of acquired_lot lists, indexed by year)
        year: int = 0
        acquired_lot_list: List[InTransaction] = []
        acquired_lot_avl: AVLTree[str, AcquiredLotAndIndex] = AVLTree()
        index: int = 0
        try:
            while True:
                acquired_lot: InTransaction = next(acquired_lot_iterator)
                if year != acquired_lot.timestamp.year:
                    if year == 0:
                        self.__min_acquired_lot_year = acquired_lot.timestamp.year
                    else:
                        self._store_acquired_lots_for_year(year, acquired_lot_list, acquired_lot_avl)
                    year = acquired_lot.timestamp.year
                    acquired_lot_list = []
                    acquired_lot_avl = AVLTree()
                    index = 0
                acquired_lot_list.append(acquired_lot)
                # Key is <timestamp>_<unique_id>
                acquired_lot_avl.insert_node(
                    f"{self._get_avl_node_key(acquired_lot.timestamp, acquired_lot.unique_id)}", AcquiredLotAndIndex(acquired_lot, index)
                )
                index += 1
        except StopIteration:
            # End of acquired_lots
            pass
        self._store_acquired_lots_for_year(year, acquired_lot_list, acquired_lot_avl)

    # AVL tree node keys have this format: <timestamp>_<unique_id>. The unique_id part is needed to disambiguate transactions
    # that have the same timestamp. Timestamp is in format "YYYYmmddHHMMSS.ffffff" and unique_id is padded right in a string of fixed
    # length (KEY_DISAMBIGUATOR_LENGTH).
    def _get_avl_node_key(self, timestamp: datetime, unique_id: str) -> str:
        return f"{timestamp.strftime('%Y%m%d%H%M%S.%f')}_{unique_id:0>{self.KEY_DISAMBIGUATOR_LENGTH}}"

    # This function calls _get_avl_node_key with unique_id=MAX_KEY_DISAMBIGUATOR, so that the generated key is larger than any other key
    # with the same timestamp.
    def _get_avl_node_key_with_max_disambiguator(self, timestamp: datetime) -> str:
        return self._get_avl_node_key(timestamp, self.MAX_KEY_DISAMBIGUATOR)

    def _store_acquired_lots_for_year(self, year: int, acquired_lot_list: List[InTransaction], acquired_lot_avl: AVLTree[str, AcquiredLotAndIndex]) -> None:
        self.__year_2_acquired_lot_list[year] = acquired_lot_list
        self.__year_2_acquired_lot_avl[year] = acquired_lot_avl

        # Acquired lots that have the same timestamp should all have the same acquired_lots list index (corresponding to the last element
        # with the same timestamp)
        previous_synchronous_acquired_lots: List[AcquiredLotAndIndex] = []
        for index, acquired_lot in enumerate(acquired_lot_list):
            if not previous_synchronous_acquired_lots or previous_synchronous_acquired_lots[0].acquired_lot.timestamp == acquired_lot.timestamp:
                previous_synchronous_acquired_lots.append(AcquiredLotAndIndex(acquired_lot, index))
            else:
                synchronous_index: int = previous_synchronous_acquired_lots[-1].index
                for previous_acquired_lot in previous_synchronous_acquired_lots:
                    previous_acquired_lot.index = synchronous_index
                previous_synchronous_acquired_lots = []

    def get_next_taxable_event_and_amount(
        self,
        taxable_event: Optional[AbstractTransaction],
        acquired_lot: Optional[InTransaction],
        taxable_event_amount: RP2Decimal,
        acquired_lot_amount: RP2Decimal,
    ) -> TaxableEventAndAcquiredLot:
        new_acquired_lot: Optional[InTransaction] = acquired_lot
        new_acquired_lot_amount: RP2Decimal = acquired_lot_amount - taxable_event_amount if acquired_lot is not None else ZERO

        try:
            new_taxable_event: AbstractTransaction = next(self.__taxable_event_iterator)
        except StopIteration:
            raise TaxableEventsExhaustedException() from None
        new_taxable_event_amount: RP2Decimal = new_taxable_event.crypto_balance_change

        # If the new taxable event has different year than the acquired lot (and it's not earn-typed), also get a new acquired lot from the new year
        if taxable_event and taxable_event.timestamp < new_taxable_event.timestamp:
            if acquired_lot:
                # Cache old-year acquired_lot amount
                self.__acquired_lot_2_partial_amount[acquired_lot] = new_acquired_lot_amount
            (_, new_acquired_lot, _, new_acquired_lot_amount) = self.get_acquired_lot_for_taxable_event(
                new_taxable_event, acquired_lot, new_taxable_event_amount, new_acquired_lot_amount
            )

        return TaxableEventAndAcquiredLot(
            taxable_event=new_taxable_event,
            acquired_lot=new_acquired_lot,
            taxable_event_amount=new_taxable_event_amount,
            acquired_lot_amount=new_acquired_lot_amount,
        )

    # After selecting the taxable event, RP2 calls this function to find the acquired_lot to pair with it. This means that the taxable
    # event can be passed to this function (which is useful for certain accounting methods)
    def get_acquired_lot_for_taxable_event(
        self, taxable_event: AbstractTransaction, acquired_lot: Optional[InTransaction], taxable_event_amount: RP2Decimal, acquired_lot_amount: RP2Decimal
    ) -> TaxableEventAndAcquiredLot:
        new_taxable_event_amount: RP2Decimal = taxable_event_amount - acquired_lot_amount
        year: int = taxable_event.timestamp.year
        while year >= self.__min_acquired_lot_year:
            if year not in self.__year_2_acquired_lot_list:
                year -= 1
                continue
            acquired_lot_list: List[InTransaction] = self.__year_2_acquired_lot_list[year]
            acquired_lot_avl: AVLTree[str, AcquiredLotAndIndex] = self.__year_2_acquired_lot_avl[year]
            first_lot: Optional[AcquiredLotAndIndex] = acquired_lot_avl.find_max_value_less_than(
                f"{self._get_avl_node_key_with_max_disambiguator(taxable_event.timestamp)}"
            )
            if first_lot is not None:
                acquired_lot_index: int = first_lot.index
                if first_lot.acquired_lot != acquired_lot_list[acquired_lot_index]:
                    raise Exception("Internal error: acquired_lot incongruence in LIFO accounting logic")
                acquired_lot_and_amount: Optional[AcquiredLotAndAmount] = self.seek_acquired_lot(reversed(acquired_lot_list[: acquired_lot_index + 1]))
                if acquired_lot_and_amount:
                    return TaxableEventAndAcquiredLot(
                        taxable_event=taxable_event,
                        acquired_lot=acquired_lot_and_amount.acquired_lot,
                        taxable_event_amount=new_taxable_event_amount,
                        acquired_lot_amount=acquired_lot_and_amount.amount,
                    )
            year -= 1

        raise AcquiredLotsExhaustedException()

    def seek_acquired_lot(self, acquired_lot_iterator: Iterator[InTransaction]) -> Optional[AcquiredLotAndAmount]:
        # This while loop make the algorithm's complexity O(nm), where n is the number of taxable events and m is
        # the number of acquired lots): for every taxable event, loop over the acquired lot list. There are non-trivial ways
        # of making this faster (by changing the data structures).
        try:
            while True:
                acquired_lot: InTransaction = next(acquired_lot_iterator)
                acquired_lot_amount: RP2Decimal
                if self._has_partial_amount(acquired_lot):
                    if self._get_partial_amount(acquired_lot) > ZERO:
                        acquired_lot_amount = self._get_partial_amount(acquired_lot)
                        self._clear_partial_amount(acquired_lot)
                        return AcquiredLotAndAmount(
                            acquired_lot=acquired_lot,
                            amount=acquired_lot_amount,
                        )
                else:
                    acquired_lot_amount = acquired_lot.crypto_in
                    self._clear_partial_amount(acquired_lot)
                    return AcquiredLotAndAmount(
                        acquired_lot=acquired_lot,
                        amount=acquired_lot_amount,
                    )
        except StopIteration:
            # End of acquired_lots
            pass

        return None

    def _has_partial_amount(self, acquired_lot: InTransaction) -> bool:
        return acquired_lot in self.__acquired_lot_2_partial_amount

    def _get_partial_amount(self, acquired_lot: InTransaction) -> RP2Decimal:
        if not self._has_partial_amount(acquired_lot):
            raise Exception(f"Internal error: acquired lot has no partial amount: {acquired_lot}")
        return self.__acquired_lot_2_partial_amount[acquired_lot]

    def _clear_partial_amount(self, acquired_lot: InTransaction) -> None:
        self.__acquired_lot_2_partial_amount[acquired_lot] = ZERO

    def validate_acquired_lot_ancestor_timestamp(self, acquired_lot: InTransaction, acquired_lot_parent: InTransaction) -> bool:
        # In LIFO the acquired_lot chain can have non-monotonic timestamps, so no validation is possible. Returning True means the validation never fails.
        return True



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from typing import Dict, List, NamedTuple, Optional, Type

from dateutil.parser import parse
from rp2.configuration import Configuration
from rp2.entry_types import TransactionType
from rp2.in_transaction import InTransaction
from rp2.input_data import InputData
from rp2.intra_transaction import IntraTransaction
from rp2.ods_parser import open_ods, parse_ods
from rp2.out_transaction import OutTransaction
from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2Error, RP2TypeError, RP2ValueError
from rp2.transaction_set import TransactionSet


class ErrorAndMessage(NamedTuple):
    error_class: Type[RP2Error]
    message: str


class TestInputParser(unittest.TestCase):
    _good_input_configuration: Configuration
    _bad_input_configuration: Configuration

    @classmethod
    def setUpClass(cls) -> None:
        TestInputParser._good_input_configuration = Configuration("./config/test_data.config", US())
        TestInputParser._bad_input_configuration = Configuration("./config/test_bad_data.config", US())

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_good_input(self) -> None:
        self._verify_good_sheet("B1", out_empty=True, intra_empty=True)
        self._verify_good_sheet("B2", out_empty=False, intra_empty=True)
        self._verify_good_sheet("B3", out_empty=True, intra_empty=False)
        self._verify_good_sheet("B4", out_empty=False, intra_empty=False)

    def _verify_good_sheet(self, sheet_name: str, out_empty: bool, intra_empty: bool) -> None:

        asset = sheet_name
        input_file_handle: object = open_ods(configuration=self._good_input_configuration, input_file_path="./input/test_data.ods")
        input_data: InputData = parse_ods(self._good_input_configuration, asset, input_file_handle)

        # In table is always present
        self._verify_non_empty_in_table(input_data.unfiltered_in_transaction_set, asset)
        if out_empty:
            self._verify_empty_table(input_data.unfiltered_out_transaction_set)
        else:
            self._verify_non_empty_out_table(input_data.unfiltered_out_transaction_set, asset)
        if intra_empty:
            self._verify_empty_table(input_data.unfiltered_intra_transaction_set)
        else:
            self._verify_non_empty_intra_table(input_data.unfiltered_intra_transaction_set, asset)

    def _verify_empty_table(self, transaction_set: TransactionSet) -> None:
        self.assertTrue(transaction_set.is_empty())

    def _verify_non_empty_in_table(self, in_transaction_set: TransactionSet, asset: str) -> None:
        unique_ids: List[str] = ["4", "3", "7", "6", "5"]
        timestamps: List[str] = [
            "2020-01-01T08:41Z",
            "2020-02-01T11:18Z",
            "2020-03-01T09:45Z",
            "2020-04-01T09:45Z",
            "2020-05-01T14:03Z",
        ]
        transaction_types: List[TransactionType] = [
            TransactionType.BUY,
            TransactionType.INTEREST,
            TransactionType.INTEREST,
            TransactionType.BUY,
            TransactionType.BUY,
        ]
        spot_prices: List[RP2Decimal] = [RP2Decimal(s) for s in ["11000", "12000", "13000", "14000", "15000"]]
        fiat_taxable_amounts: List[RP2Decimal] = [RP2Decimal(s) for s in ["0", "24000", "39000", "0", "0"]]
        crypto_balance_changes: List[RP2Decimal] = [RP2Decimal(s) for s in ["1", "2", "3", "4", "5"]]
        fiat_balance_changes: List[RP2Decimal] = [RP2Decimal(s) for s in ["11100", "24000", "39000", "56400", "75500"]]
        is_taxable_values: List[bool] = [False, True, True, False, False]

        count: int = 0
        unique_id: str
        timestamp: str
        transaction_type: TransactionType
        spot_price: RP2Decimal
        fiat_taxable_amount: RP2Decimal
        crypto_balance_change: RP2Decimal
        fiat_balance_change: RP2Decimal
        is_taxable: bool

        transaction: Optional[InTransaction] = None
        previous_transaction: Optional[InTransaction] = None
        for (  # type: ignore
            transaction,
            unique_id,
            timestamp,
            transaction_type,
            spot_price,
            fiat_taxable_amount,
            crypto_balance_change,
            fiat_balance_change,
            is_taxable,
        ) in zip(  # type: ignore
            in_transaction_set,
            unique_ids,
            timestamps,
            transaction_types,
            spot_prices,
            fiat_taxable_amounts,
            crypto_balance_changes,
            fiat_balance_changes,
            is_taxable_values,
        ):
            if not in_transaction_set or not transaction:  # Unwrap the Optional types to keep mypy happy
                raise Exception("Internal error: in_transaction_set or transaction are None")
            self.assertEqual(in_transaction_set.get_parent(transaction), previous_transaction)
            self.assertEqual(transaction.unique_id, unique_id)
            self.assertEqual(transaction.timestamp, parse(timestamp))
            self.assertEqual(transaction.transaction_type, transaction_type)
            self.assertEqual(transaction.spot_price, spot_price)
            self.assertEqual(transaction.asset, asset)
            self.assertEqual(transaction.fiat_taxable_amount, fiat_taxable_amount)
            self.assertEqual(transaction.crypto_balance_change, crypto_balance_change)
            self.assertEqual(transaction.fiat_balance_change, fiat_balance_change)
            self.assertEqual(transaction.is_taxable(), is_taxable)
            previous_transaction = transaction
            count += 1
        self.assertEqual(count, 5)

    def _verify_non_empty_out_table(self, out_transaction_set: TransactionSet, asset: str) -> None:
        unique_ids: List[str] = ["16", "15", "17", "13", "14"]
        timestamps: List[str] = [
            "2020-01-11T11:15Z",
            "2020-02-11T19:58Z",
            "2020-04-11T07:10Z",
            "2020-04-12T17:50Z",
            "2021-06-11T05:31Z",
        ]
        transaction_types: List[TransactionType] = [
            TransactionType.SELL,
            TransactionType.SELL,
            TransactionType.GIFT,
            TransactionType.DONATE,
            TransactionType.SELL,
        ]
        spot_prices: List[RP2Decimal] = [RP2Decimal(d) for d in ["11200", "12200", "14200", "14300", "20200"]]
        fiat_taxable_amounts: List[RP2Decimal] = [RP2Decimal(d) for d in ["2240", "12200", "71000", "54197", "40400.0"]]
        crypto_balance_changes: List[RP2Decimal] = [RP2Decimal(d) for d in ["0.2", "1", "5", "3.79", "2.01"]]
        fiat_balance_changes: List[RP2Decimal] = [RP2Decimal(d) for d in ["2240", "12200", "71000", "54197", "40602.0"]]
        is_taxable_values: List[bool] = [True, True, True, True, True]

        count: int = 0
        unique_id: str
        timestamp: str
        transaction_type: TransactionType
        spot_price: RP2Decimal
        fiat_taxable_amount: RP2Decimal
        crypto_balance_change: RP2Decimal
        fiat_balance_change: RP2Decimal
        is_taxable: bool

        transaction: Optional[OutTransaction] = None
        previous_transaction: Optional[OutTransaction] = None

        for (  # type: ignore
            transaction,
            unique_id,
            timestamp,
            transaction_type,
            spot_price,
            fiat_taxable_amount,
            crypto_balance_change,
            fiat_balance_change,
            is_taxable,
        ) in zip(  # type: ignore
            out_transaction_set,
            unique_ids,
            timestamps,
            transaction_types,
            spot_prices,
            fiat_taxable_amounts,
            crypto_balance_changes,
            fiat_balance_changes,
            is_taxable_values,
        ):
            if not out_transaction_set or not transaction:  # Unwrap the Optional types to keep mypy happy
                raise Exception("Internal error: in_transaction_set or transaction are None")
            self.assertEqual(out_transaction_set.get_parent(transaction), previous_transaction)
            self.assertEqual(transaction.unique_id, unique_id)
            self.assertEqual(transaction.timestamp, parse(timestamp))
            self.assertEqual(transaction.transaction_type, transaction_type)
            self.assertEqual(transaction.spot_price, spot_price)
            self.assertEqual(transaction.asset, asset)
            self.assertEqual(transaction.fiat_taxable_amount, fiat_taxable_amount)
            self.assertEqual(transaction.crypto_balance_change, crypto_balance_change)
            self.assertEqual(transaction.fiat_balance_change, fiat_balance_change)
            self.assertEqual(transaction.is_taxable(), is_taxable)
            previous_transaction = transaction
            count += 1
        self.assertEqual(count, 5)

    def _verify_non_empty_intra_table(self, intra_transaction_set: TransactionSet, asset: str) -> None:
        unique_ids: List[str] = ["26", "24", "25", "23"]
        timestamps: List[str] = [
            "2020-01-21T18:33:14.342Z",
            "2020-02-21T20:23:31Z",
            "2020-05-21T12:58:10Z",
            "2021-07-21T10:02:02Z",
        ]
        transaction_types: List[TransactionType] = [
            TransactionType.MOVE,
            TransactionType.MOVE,
            TransactionType.MOVE,
            TransactionType.MOVE,
        ]
        spot_prices: List[RP2Decimal] = [RP2Decimal(d) for d in ["11400", "0", "14400", "21400"]]
        fiat_taxable_amounts: List[RP2Decimal] = [RP2Decimal(d) for d in ["114", "0", "288", "856"]]
        crypto_balance_changes: List[RP2Decimal] = [RP2Decimal(d) for d in ["0.01", "0", "0.02", "0.04"]]
        fiat_balance_changes: List[RP2Decimal] = [RP2Decimal(d) for d in ["114", "0", "288", "856"]]
        is_taxable_values: List[bool] = [True, False, True, True]

        count: int = 0
        unique_id: str
        timestamp: str
        transaction_type: TransactionType
        spot_price: RP2Decimal
        fiat_taxable_amount: RP2Decimal
        crypto_balance_change: RP2Decimal
        fiat_balance_change: RP2Decimal
        is_taxable: bool

        transaction: Optional[IntraTransaction] = None
        previous_transaction: Optional[IntraTransaction] = None
        for (  # type: ignore
            transaction,
            unique_id,
            timestamp,
            transaction_type,
            spot_price,
            fiat_taxable_amount,
            crypto_balance_change,
            fiat_balance_change,
            is_taxable,
        ) in zip(  # type: ignore
            intra_transaction_set,
            unique_ids,
            timestamps,
            transaction_types,
            spot_prices,
            fiat_taxable_amounts,
            crypto_balance_changes,
            fiat_balance_changes,
            is_taxable_values,
        ):
            if not intra_transaction_set or not transaction:  # Unwrap the Optional types to keep mypy happy
                raise Exception("Internal error: intra_transaction_set or transaction are None")
            self.assertEqual(intra_transaction_set.get_parent(transaction), previous_transaction)
            self.assertEqual(transaction.unique_id, unique_id)
            self.assertEqual(transaction.timestamp, parse(timestamp))
            self.assertEqual(transaction.transaction_type, transaction_type)
            self.assertEqual(transaction.spot_price, spot_price)
            self.assertEqual(transaction.asset, asset)
            self.assertEqual(transaction.fiat_taxable_amount, fiat_taxable_amount)
            self.assertEqual(transaction.crypto_balance_change, crypto_balance_change)
            self.assertEqual(transaction.fiat_balance_change, fiat_balance_change)
            self.assertEqual(transaction.is_taxable(), is_taxable)
            previous_transaction = transaction
            count += 1
        self.assertEqual(count, 4)

    def test_bad_input(self) -> None:

        sheets_to_expected_messages: Dict[str, ErrorAndMessage] = {
            "B1": ErrorAndMessage(RP2ValueError, "IN table not found"),
            "B2": ErrorAndMessage(RP2ValueError, 'Found an invalid cell "foo" while looking for a table-begin token'),
            "B3": ErrorAndMessage(RP2ValueError, 'Found an invalid cell "bar" while looking for a table-begin token'),
            "B4": ErrorAndMessage(RP2ValueError, "Found end-table keyword without having found a table-begin keyword first"),
            "B5": ErrorAndMessage(RP2ValueError, "Found end-table keyword without having found a table-begin keyword first"),
            "B6": ErrorAndMessage(RP2ValueError, "TABLE END not found for EntrySetType.IN table"),
            "B7": ErrorAndMessage(RP2ValueError, "TABLE END not found for EntrySetType.OUT table"),
            "B8": ErrorAndMessage(RP2ValueError, "TABLE END not found for EntrySetType.INTRA table"),
            "B9": ErrorAndMessage(RP2ValueError, 'Found "IN" keyword while parsing table EntrySetType.IN'),
            "B10": ErrorAndMessage(RP2ValueError, 'Found "OUT" keyword while parsing table EntrySetType.OUT'),
            "B11": ErrorAndMessage(RP2ValueError, 'Found "INTRA" keyword while parsing table EntrySetType.INTRA'),
            "B12": ErrorAndMessage(RP2ValueError, 'Found "OUT" keyword while parsing table EntrySetType.IN'),
            "B13": ErrorAndMessage(RP2ValueError, 'Found "OUT" keyword while parsing table EntrySetType.INTRA'),
            "B14": ErrorAndMessage(RP2ValueError, "IN table not found or empty"),
            "B15": ErrorAndMessage(RP2ValueError, "TABLE END not found for EntrySetType.IN table"),
            "B16": ErrorAndMessage(RP2ValueError, "Parameter 'data' has length .*, but required minimum from in-table headers in .* is .*"),
            "B17": ErrorAndMessage(RP2ValueError, "Parameter 'data' has length .*, but required minimum from out-table headers in .* is .*"),
            "B18": ErrorAndMessage(RP2TypeError, "Parameter 'asset' has non-string value .*"),
            "B19": ErrorAndMessage(RP2ValueError, "Found an empty cell while parsing table EntrySetType.OUT"),
            "B20": ErrorAndMessage(RP2ValueError, "IN table not found"),
            "B21": ErrorAndMessage(RP2ValueError, "IN table not found or empty"),
            "B22": ErrorAndMessage(RP2ValueError, "Found data with no header"),
            "B23": ErrorAndMessage(RP2ValueError, "Parameter 'timestamp' value has no timezone info: .*"),
            "B24": ErrorAndMessage(RP2ValueError, "Parameter 'exchange' value is not known: .*"),
            "B25": ErrorAndMessage(RP2ValueError, "Parameter 'holder' value is not known: .*"),
            "B26": ErrorAndMessage(RP2ValueError, "Parameter 'transaction_type' has invalid transaction type value: .*"),
            "B27": ErrorAndMessage(RP2ValueError, "Parameter 'asset' value is not known: .*"),
            "B28": ErrorAndMessage(RP2ValueError, "Parameter 'crypto_in' has non-positive value .*"),
            "B29": ErrorAndMessage(RP2ValueError, "Parameter 'spot_price' has non-positive value .*"),
            "B30": ErrorAndMessage(RP2ValueError, "Parameter 'fiat_fee' has non-positive value .*"),
            "B31": ErrorAndMessage(RP2ValueError, "Found an empty cell while parsing table EntrySetType.IN"),
            "B32": ErrorAndMessage(RP2ValueError, "TABLE END not found for EntrySetType.OUT table"),
            "B33": ErrorAndMessage(RP2ValueError, "Found more than one IN symbol"),
            "B34": ErrorAndMessage(RP2ValueError, "IN table not found or empty"),
            "B35": ErrorAndMessage(RP2ValueError, "IN table not found or empty"),
            "B36": ErrorAndMessage(RP2ValueError, "IN table not found or empty"),
        }

        sheet: str
        message: str
        for sheet, (error_class, message) in sheets_to_expected_messages.items():
            with self.assertRaisesRegex(error_class, message):
                asset: str = sheet
                input_file_handle: object = open_ods(configuration=self._bad_input_configuration, input_file_path="./input/test_bad_data.ods")
                parse_ods(self._bad_input_configuration, asset, input_file_handle)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cProfile
from datetime import date
import os
import sys
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from importlib import import_module
from pathlib import Path
from pkgutil import iter_modules
from types import ModuleType
from typing import Dict, List

from rp2.abstract_accounting_method import AbstractAccountingMethod
from rp2.abstract_country import AbstractCountry
from rp2.abstract_report_generator import AbstractReportGenerator
from rp2.computed_data import ComputedData
from rp2.configuration import MAX_DATE, MIN_DATE, VERSION, Configuration
from rp2.input_data import InputData
from rp2.logger import LOG_FILE, LOGGER
from rp2.ods_parser import open_ods, parse_ods
from rp2.tax_engine import compute_tax

_ACCOUNTING_METHOD_PACKAGE = "rp2.plugin.accounting_method"
_REPORT_GENERATOR_PACKAGE = "rp2.plugin.report"


def rp2_main(country: AbstractCountry) -> None:
    if "RP2_ENABLE_PROFILER" in os.environ:
        cProfile.runctx("_rp2_main_internal(country)", globals(), locals())
    else:
        _rp2_main_internal(country)


def _rp2_main_internal(country: AbstractCountry) -> None:
    args: Namespace
    assets: List[str]
    parser: ArgumentParser

    AbstractCountry.type_check("country", country)

    accounting_methods: List[str] = _find_accounting_methods()

    parser = _setup_argument_parser(accounting_methods)
    args = parser.parse_args()

    _setup_paths(parser=parser, configuration_file=args.configuration_file, input_file=args.input_file, output_dir=args.output_dir)

    try:
        LOGGER.info("Country: %s", country.country_iso_code)

        accounting_method_module: ModuleType = import_module(f"{_ACCOUNTING_METHOD_PACKAGE}.{args.method}", package=_ACCOUNTING_METHOD_PACKAGE)
        if not hasattr(accounting_method_module, "AccountingMethod"):
            LOGGER.error("Accounting method plugin %s doesn't have an AccountingMethod class")
            sys.exit(1)
        accounting_method: AbstractAccountingMethod = accounting_method_module.AccountingMethod()
        LOGGER.info("Accounting Method: %s", args.method)

        configuration: Configuration = Configuration(
            configuration_path=args.configuration_file, country=country, from_date=args.from_date, to_date=args.to_date
        )
        LOGGER.info("Configuration file: %s", args.configuration_file)
        LOGGER.debug("Configuration object: %s", configuration)

        if args.asset:
            assets = [args.asset]
        else:
            assets = list(configuration.assets)
        assets.sort()

        asset_to_computed_data: Dict[str, ComputedData] = {}
        asset: str

        LOGGER.info("Input file: %s", args.input_file)
        input_file_handle: object = open_ods(configuration=configuration, input_file_path=args.input_file)
        for asset in assets:
            LOGGER.info("Processing %s", asset)

            input_data: InputData = parse_ods(configuration=configuration, asset=asset, input_file_handle=input_file_handle)
            LOGGER.debug("InputData object: %s", input_data)

            computed_data: ComputedData = compute_tax(configuration=configuration, accounting_method=accounting_method, input_data=input_data)
            LOGGER.debug("ComputedData object: %s", computed_data)

            asset_to_computed_data[asset] = computed_data

        # Run non-country-specific report generators
        _find_and_run_report_generators(
            package_path=_REPORT_GENERATOR_PACKAGE,
            args=args,
            country=country,
            accounting_method=accounting_method,
            asset_to_computed_data=asset_to_computed_data,
            from_date=configuration.from_date,
            to_date=configuration.to_date,
        )
        # Run country-specific report generators
        _find_and_run_report_generators(
            package_path=f"{_REPORT_GENERATOR_PACKAGE}.{country.country_iso_code}",
            args=args,
            country=country,
            accounting_method=accounting_method,
            asset_to_computed_data=asset_to_computed_data,
            from_date=configuration.from_date,
            to_date=configuration.to_date,
        )
    except Exception:  # pylint: disable=broad-except
        LOGGER.exception("Fatal exception occurred:")

    LOGGER.info("Log file: %s", LOG_FILE)
    LOGGER.info("Generated output directory: %s", args.output_dir)
    LOGGER.info("Done")


def _find_and_run_report_generators(
    package_path: str,
    args: Namespace,
    country: AbstractCountry,
    accounting_method: AbstractAccountingMethod,
    asset_to_computed_data: Dict[str, ComputedData],
    from_date: date,
    to_date: date,
) -> None:
    # Load report generator plugins and call their generate() method
    package: ModuleType = import_module(package_path)
    plugin_name: str
    is_package: bool
    package_found: bool = False
    for *_, plugin_name, is_package in iter_modules(package.__path__, package.__name__ + "."):
        if is_package:
            continue
        if args.plugin and plugin_name != f"{_REPORT_GENERATOR_PACKAGE}.{args.plugin}":
            continue
        output_module: ModuleType = import_module(plugin_name, package=_REPORT_GENERATOR_PACKAGE)
        if hasattr(output_module, "Generator"):
            generator: AbstractReportGenerator = output_module.Generator()
            LOGGER.debug("Generator object: '%s'", generator)
            LOGGER.info("Generating output for plugin '%s'", plugin_name)
            if not hasattr(generator, "generate"):
                LOGGER.error("Plugin '%s' has no 'generate' method. Exiting...", plugin_name)
                sys.exit(1)
            generator.generate(
                country=country,
                accounting_method=repr(accounting_method),
                asset_to_computed_data=asset_to_computed_data,
                output_dir_path=args.output_dir,
                output_file_prefix=args.prefix,
                from_date=from_date,
                to_date=to_date,
            )
        package_found = True

    if not package_found:
        if args.plugin:
            LOGGER.error("Report Generator plugin '%s' not found. Exiting...", args.plugin)
        else:
            LOGGER.error("No report generator plugin found. Exiting...")
        sys.exit(1)


def _find_accounting_methods() -> List[str]:
    # Load accounting method plugins
    package: ModuleType = import_module(_ACCOUNTING_METHOD_PACKAGE)
    plugin_name: str
    is_package: bool
    result: List[str] = []

    for *_, plugin_name, is_package in iter_modules(package.__path__, package.__name__ + "."):
        if is_package:
            continue
        result.append(plugin_name.rsplit(".", 1)[1])

    if not result:
        LOGGER.error("No accounting method plugins found. Exiting...")

    return result


def _setup_argument_parser(accounting_methods: List[str]) -> ArgumentParser:
    parser: ArgumentParser = ArgumentParser(
        description=(
            "Generate capital gain/loss report and balances for crypto holdings. Links:\n"
            "- documentation: https://github.com/eprbell/rp2/blob/main/README.md\n"
            "- FAQ: https://github.com/eprbell/rp2/blob/main/docs/user_faq.md\n"
            "- donations: https://eprbell.github.io/eprbell/donations.html\n"
            "- leave a star on Github: https://github.com/eprbell/rp2"
        ),
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "-a",
        "--asset",
        action="store",
        help="Generate report only for the given ASSET",
        metavar="ASSET",
        type=str,
    )
    parser.add_argument(
        "-f",
        "--from_date",
        action="store",
        default=MIN_DATE,
        help="Generate report from the given date (in ISO 8601 format: e.g. YYYY-MM-DD)",
        metavar="DATE",
        type=date.fromisoformat,
    )
    parser.add_argument(
        "-l",
        "--plugin",
        action="store",
        help="Generate report only using the given PLUGIN",
        metavar="PLUGIN",
        type=str,
    )
    parser.add_argument(
        "-m",
        "--method",
        default="fifo",
        choices=accounting_methods,
        help=f"accounting method (default: '%(default)s'). Supported values: {', '.join(accounting_methods)}",
        metavar="METHOD",
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        action="store",
        default="output/",
        help="Write output to OUTPUT_DIR  (default: '%(default)s')",
        metavar="OUTPUT_DIR",
        type=str,
    )
    parser.add_argument(
        "-p",
        "--prefix",
        action="store",
        default="",
        help="Prepend output file names with PREFIX",
        metavar="PREFIX",
        type=str,
    )
    parser.add_argument(
        "-t",
        "--to_date",
        action="store",
        default=MAX_DATE,
        help="Generate report up to the given date (in ISO 8601 format: e.g. YYYY-MM-DD)",
        metavar="DATE",
        type=date.fromisoformat,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"RP2 {VERSION} (https://pypi.org/project/rp2/)",
        help="Print RP2 version",
    )
    parser.add_argument(
        "configuration_file",
        action="store",
        help="Configuration file",
        metavar="CONFIGURATION",
        type=str,
    )
    parser.add_argument(
        "input_file",
        action="store",
        help="ODS file containing input transactions",
        metavar="INPUT",
        type=str,
    )

    return parser


def _setup_paths(parser: ArgumentParser, configuration_file: str, input_file: str, output_dir: str) -> None:
    if not Path(configuration_file).exists():
        print(f"Configuration file '{configuration_file}' not found")
        parser.print_help()
        sys.exit(1)

    if not input_file.endswith(".ods"):
        print(f"Input file '{input_file}' does not end with '.ods'")
        parser.print_help()
        sys.exit(1)

    if not Path(input_file).exists():
        print(f"Input file '{input_file}' not found")
        parser.print_help()
        sys.exit(1)

    output_dir_path: Path = Path(output_dir)
    if not output_dir_path.exists():
        output_dir_path.mkdir(parents=True)
    if not output_dir_path.is_dir():
        print(f"output_dir '{output_dir}' exists but it's not a directory")
        parser.print_help()
        sys.exit(1)



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import unittest

from dateutil.tz import tzoffset
from rp2.configuration import Configuration
from rp2.entry_types import TransactionType
from rp2.intra_transaction import IntraTransaction
from rp2.out_transaction import OutTransaction
from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError


class TestOutTransaction(unittest.TestCase):
    _configuration: Configuration

    @classmethod
    def setUpClass(cls) -> None:
        TestOutTransaction._configuration = Configuration("./config/test_data.config", US())

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_taxable_out_transaction(self) -> None:
        out_transaction: OutTransaction = OutTransaction(
            self._configuration,
            "6/1/2020 3:59:59 -04:00",
            "B1",
            "Coinbase Pro",
            "Bob",
            "SELL",
            RP2Decimal("900.9"),
            RP2Decimal("2.2"),
            RP2Decimal("0.01"),
            unique_id=38,
        )
        OutTransaction.type_check("my_instance", out_transaction)

        self.assertTrue(out_transaction.is_taxable())
        self.assertEqual(RP2Decimal("1981.98"), out_transaction.fiat_taxable_amount)
        self.assertEqual(RP2Decimal("2.2"), out_transaction.crypto_taxable_amount)
        self.assertEqual("38", out_transaction.unique_id)
        self.assertEqual(2020, out_transaction.timestamp.year)
        self.assertEqual(6, out_transaction.timestamp.month)
        self.assertEqual(1, out_transaction.timestamp.day)
        self.assertEqual(3, out_transaction.timestamp.hour)
        self.assertEqual(59, out_transaction.timestamp.minute)
        self.assertEqual(59, out_transaction.timestamp.second)
        self.assertEqual(tzoffset(None, -14400), out_transaction.timestamp.tzinfo)
        self.assertEqual("B1", out_transaction.asset)
        self.assertEqual("Coinbase Pro", out_transaction.exchange)
        self.assertEqual("Bob", out_transaction.holder)
        self.assertEqual(TransactionType.SELL, out_transaction.transaction_type)
        self.assertEqual(RP2Decimal("900.9"), out_transaction.spot_price)
        self.assertEqual(RP2Decimal("2.2"), out_transaction.crypto_out_no_fee)
        self.assertEqual(RP2Decimal("2.21"), out_transaction.crypto_balance_change)
        self.assertEqual(RP2Decimal("1990.989"), out_transaction.fiat_balance_change)

        self.assertEqual(
            str(out_transaction),
            """OutTransaction:
  id=38
  timestamp=2020-06-01 03:59:59.000000 -0400
  asset=B1
  exchange=Coinbase Pro
  holder=Bob
  transaction_type=TransactionType.SELL
  spot_price=900.9000
  crypto_out_no_fee=2.20000000
  crypto_fee=0.01000000
  is_taxable=True
  fiat_taxable_amount=1981.9800""",
        )
        self.assertEqual(
            out_transaction.to_string(2, repr_format=False, extra_data=["foobar", "qwerty"]),
            """    OutTransaction:
      id=38
      timestamp=2020-06-01 03:59:59.000000 -0400
      asset=B1
      exchange=Coinbase Pro
      holder=Bob
      transaction_type=TransactionType.SELL
      spot_price=900.9000
      crypto_out_no_fee=2.20000000
      crypto_fee=0.01000000
      is_taxable=True
      fiat_taxable_amount=1981.9800
      foobar
      qwerty""",
        )
        self.assertEqual(
            out_transaction.to_string(2, repr_format=True, extra_data=["foobar", "qwerty"]),
            (
                "    OutTransaction("
                "id='38', "
                "timestamp='2020-06-01 03:59:59.000000 -0400', "
                "asset='B1', "
                "exchange='Coinbase Pro', "
                "holder='Bob', "
                "transaction_type=<TransactionType.SELL: 'sell'>, "
                "spot_price=900.9000, "
                "crypto_out_no_fee=2.20000000, "
                "crypto_fee=0.01000000, "
                "is_taxable=True, "
                "fiat_taxable_amount=1981.9800, "
                "foobar, "
                "qwerty)"
            ),
        )

    def test_out_transaction_equality_and_hashing(self) -> None:
        out_transaction: OutTransaction = OutTransaction(
            self._configuration,
            "6/1/2020 3:59:59 -04:00",
            "B1",
            "Coinbase Pro",
            "Bob",
            "SELL",
            RP2Decimal("900.9"),
            RP2Decimal("2.2"),
            RP2Decimal("0.01"),
            unique_id=38,
        )
        out_transaction2: OutTransaction = OutTransaction(
            self._configuration,
            "6/1/2020 3:59:59 -04:00",
            "B1",
            "Coinbase Pro",
            "Bob",
            "SELL",
            RP2Decimal("900.9"),
            RP2Decimal("2.2"),
            RP2Decimal("0.01"),
            unique_id=38,
        )
        out_transaction3: OutTransaction = OutTransaction(
            self._configuration,
            "6/1/2020 3:59:59 -04:00",
            "B1",
            "Coinbase Pro",
            "Bob",
            "SELL",
            RP2Decimal("900.9"),
            RP2Decimal("2.2"),
            RP2Decimal("0.01"),
            unique_id=7,
        )
        self.assertEqual(out_transaction, out_transaction)
        self.assertEqual(out_transaction, out_transaction2)
        self.assertNotEqual(out_transaction, out_transaction3)
        self.assertEqual(hash(out_transaction), hash(out_transaction))
        self.assertEqual(hash(out_transaction), hash(out_transaction2))
        # These hashes would only be equal in case of hash collision (possible but very unlikey)
        self.assertNotEqual(hash(out_transaction), hash(out_transaction3))

    def test_bad_to_string(self) -> None:
        out_transaction: OutTransaction = OutTransaction(
            self._configuration,
            "6/1/2020 3:59:59 -04:00",
            "B1",
            "Coinbase Pro",
            "Bob",
            "SELL",
            RP2Decimal("900.9"),
            RP2Decimal("2.2"),
            RP2Decimal("0.01"),
            unique_id=38,
        )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'indent' has non-integer value"):
            out_transaction.to_string(None, repr_format=False, extra_data=["foobar", "qwerty"])  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'indent' has non-positive value.*"):
            out_transaction.to_string(-1, repr_format=False, extra_data=["foobar", "qwerty"])
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'repr_format' has non-bool value .*"):
            out_transaction.to_string(1, repr_format="False", extra_data=["foobar", "qwerty"])  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'extra_data' is not of type List"):
            out_transaction.to_string(1, repr_format=False, extra_data="foobar")  # type: ignore

    def test_bad_out_transaction(self) -> None:
        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string:.*"):
            OutTransaction.type_check(None, None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_instance' is not of type OutTransaction:.*"):
            OutTransaction.type_check("my_instance", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_instance' is not of type OutTransaction: IntraTransaction"):
            OutTransaction.type_check(
                "my_instance",
                IntraTransaction(
                    self._configuration,
                    "2021-01-12T11:51:38Z",
                    "B1",
                    "BlockFi",
                    "Bob",
                    "Coinbase",
                    "Alice",
                    RP2Decimal("10000"),
                    RP2Decimal("1"),
                    RP2Decimal("1"),
                    unique_id=45,
                ),
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            OutTransaction(
                (1, 2, 3),  # type: ignore
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "GIFT",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            OutTransaction(
                None,  # type: ignore
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'unique_id' has non-positive value .*"):
            # Bad unique_id
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "GIFT",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=-38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'unique_id' has non-integer value .*"):
            # Bad unique_id
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=(1, 2, 3),  # type: ignore
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'timestamp' has non-string value .*"):
            # Bad timestamp
            OutTransaction(
                self._configuration,
                None,  # type: ignore
                "B1",
                "Coinbase Pro",
                "Bob",
                "gIfT",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'timestamp' value has no timezone info: .*"):
            # Bad timestamp
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59",
                "B1",
                "Coinbase Pro",
                "Bob",
                "selL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'timestamp' has non-string value .*"):
            # Bad timestamp
            OutTransaction(
                self._configuration,
                (1, 2, 3),  # type: ignore
                "B1",
                "Coinbase Pro",
                "Bob",
                "GIFT",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'asset' value is not known: .*"):
            # Bad asset
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "YYY",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            # Bad asset
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                None,  # type: ignore
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'exchange' value is not known: .*"):
            # Bad exchange
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "CoinbasE Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'exchange' has non-string value .*"):
            # Bad exchange
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                None,  # type: ignore
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'holder' value is not known: .*"):
            # Bad holder
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "foobar",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'holder' has non-string value .*"):
            # Bad holder
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                None,  # type: ignore
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, ".*OutTransaction .*, id.*invalid transaction type .*"):
            # Bad transaction type
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "Buy",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, ".*OutTransaction .*, id.*invalid transaction type .*"):
            # Bad transaction type
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "iNteResT",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter .* has invalid transaction type value: .*"):
            # Bad transaction type
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BEND",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter .* has non-string value .*"):
            # Bad transaction type
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                None,  # type: ignore
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "OutTransaction .*, id.*parameter 'spot_price' cannot be 0"):
            # Bad spot price
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("0"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'spot_price' has non-positive value .*"):
            # Bad spot price
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("-900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'spot_price' has non-RP2Decimal value ,*"):
            # Bad spot price
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "GIFT",
                None,  # type: ignore
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_out_no_fee' has zero value"):
            # Bad crypto out no fee
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "GIFT",
                RP2Decimal("900.9"),
                RP2Decimal("0"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_out_no_fee' has non-positive value .*"):
            # Bad crypto out no fee
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("-2.2"),
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'crypto_out_no_fee' has non-RP2Decimal value .*"):
            # Bad crypto out no fee
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                None,  # type: ignore
                RP2Decimal("0"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_fee' has non-positive value .*"):
            # Bad crypto fee
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("-0.1"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'crypto_fee' has non-RP2Decimal value .*"):
            # Bad crypto fee
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                "foobar",  # type: ignore
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'fiat_out_no_fee' has non-positive value .*"):
            # Bad fiat_out_no_fee
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                fiat_out_no_fee=RP2Decimal("-0.1"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'fiat_out_no_fee' has non-RP2Decimal value .*"):
            # Bad fiat_out_no_fee
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                fiat_out_no_fee="foobar",  # type: ignore
                unique_id=38,
            )

        with self.assertRaisesRegex(RP2ValueError, "Parameter 'fiat_fee' has non-positive value .*"):
            # Bad fiat fee
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                fiat_out_no_fee=RP2Decimal("1800"),
                fiat_fee=RP2Decimal("-10"),
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'fiat_fee' has non-RP2Decimal value .*"):
            # Bad fiat fee
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                fiat_out_no_fee=RP2Decimal("1800"),
                fiat_fee="foobar",  # type: ignore
                unique_id=38,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'notes' has non-string value .*"):
            # Bad notes
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0"),
                notes=(1, 2, 3),  # type: ignore
                unique_id=38,
            )

        with self.assertLogs(level="WARNING") as log:
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0.1"),
                crypto_out_with_fee=RP2Decimal("2.2"),
                unique_id=38,
            )
            self.assertTrue(re.search("crypto_out_with_fee != crypto_out_no_fee.*crypto_fee:.*", log.output[0]))  # type: ignore
        with self.assertLogs(level="WARNING") as log:
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "SELL",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0.1"),
                fiat_out_no_fee=RP2Decimal("1981.98"),
                fiat_fee=RP2Decimal("5.9"),
                unique_id=38,
            )
            self.assertTrue(re.search("crypto_fee.*spot_price.*!= fiat_fee.*:.*", log.output[0]))  # type: ignore
        with self.assertLogs(level="WARNING") as log:
            OutTransaction(
                self._configuration,
                "6/1/2020 3:59:59 -04:00",
                "B1",
                "Coinbase Pro",
                "Bob",
                "GIFT",
                RP2Decimal("900.9"),
                RP2Decimal("2.2"),
                RP2Decimal("0.1"),
                fiat_out_no_fee=RP2Decimal("1081.98"),
                fiat_fee=RP2Decimal("90.09"),
                unique_id=38,
            )
            self.assertTrue(re.search("crypto_out_no_fee.*spot_price.*!= fiat_out_no_fee.*:.*", log.output[0]))  # type: ignore


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from datetime import datetime
from pathlib import Path
from typing import Optional

LOG_FILE: str = f"./log/rp2_{datetime.now().strftime('%Y_%m_%d_%H_%M_%S_%f')}.log"
Path("./log").mkdir(parents=True, exist_ok=True)

def create_logger(logger_name: str="rp2") -> logging.Logger:
    logger: logging.Logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)

    _console_handler: logging.StreamHandler = logging.StreamHandler()
    _console_format: logging.Formatter = logging.Formatter("%(levelname)s: %(message)s")
    _console_handler.setLevel(logging.INFO)
    _console_handler.setFormatter(_console_format)

    _file_handler: logging.FileHandler = logging.FileHandler(LOG_FILE)
    _file_format: logging.Formatter = logging.Formatter("%(asctime)s/%(name)s/%(levelname)s: %(message)s")
    _log_level: Optional[str] = os.environ.get("LOG_LEVEL")
    _log_level = "INFO" if not _log_level else _log_level
    _file_handler.setLevel(_log_level)
    _file_handler.setFormatter(_file_format)

    logger.addHandler(_console_handler)
    logger.addHandler(_file_handler)

    return logger

LOGGER: logging.Logger = create_logger()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from dateutil.tz import tzutc
from rp2.configuration import Configuration
from rp2.entry_types import TransactionType
from rp2.in_transaction import InTransaction
from rp2.intra_transaction import IntraTransaction
from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError


class TestIntraTransaction(unittest.TestCase):
    _configuration: Configuration

    @classmethod
    def setUpClass(cls) -> None:
        TestIntraTransaction._configuration = Configuration("./config/test_data.config", US())

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_taxable_intra_transaction(self) -> None:
        intra_transaction: IntraTransaction = IntraTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "Coinbase Pro",
            "Bob",
            "BlockFi",
            "Alice",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            RP2Decimal("1.9998"),
            unique_id=19,
        )

        IntraTransaction.type_check("my_instance", intra_transaction)
        self.assertTrue(intra_transaction.is_taxable())
        self.assertEqual(RP2Decimal("0.4"), intra_transaction.fiat_taxable_amount)
        self.assertEqual("19", intra_transaction.unique_id)
        self.assertEqual(2021, intra_transaction.timestamp.year)
        self.assertEqual(1, intra_transaction.timestamp.month)
        self.assertEqual(2, intra_transaction.timestamp.day)
        self.assertEqual(8, intra_transaction.timestamp.hour)
        self.assertEqual(42, intra_transaction.timestamp.minute)
        self.assertEqual(43, intra_transaction.timestamp.second)
        self.assertEqual(882000, intra_transaction.timestamp.microsecond)
        self.assertEqual(tzutc(), intra_transaction.timestamp.tzinfo)
        self.assertEqual("B1", intra_transaction.asset)
        self.assertEqual("Coinbase Pro", intra_transaction.from_exchange)
        self.assertEqual("Bob", intra_transaction.from_holder)
        self.assertEqual("BlockFi", intra_transaction.to_exchange)
        self.assertEqual("Alice", intra_transaction.to_holder)
        self.assertEqual(TransactionType.MOVE, intra_transaction.transaction_type)
        self.assertEqual(RP2Decimal("1000"), intra_transaction.spot_price)
        self.assertEqual(RP2Decimal("2.0002"), intra_transaction.crypto_sent)
        self.assertEqual(RP2Decimal("1.9998"), intra_transaction.crypto_received)
        self.assertEqual(RP2Decimal("0.0004"), intra_transaction.crypto_fee)
        self.assertEqual(RP2Decimal("0.4"), intra_transaction.fiat_fee)
        self.assertEqual(RP2Decimal("0.0004"), intra_transaction.crypto_balance_change)
        self.assertEqual(RP2Decimal("0.4"), intra_transaction.fiat_balance_change)

        self.assertEqual(
            str(intra_transaction),
            """IntraTransaction:
  id=19
  timestamp=2021-01-02 08:42:43.882000 +0000
  asset=B1
  from_exchange=Coinbase Pro
  from_holder=Bob
  to_exchange=BlockFi
  to_holder=Alice
  transaction_type=TransactionType.MOVE
  spot_price=1000.0000
  crypto_sent=2.00020000
  crypto_received=1.99980000
  crypto_fee=0.00040000
  fiat_fee=0.4000
  is_taxable=True
  fiat_taxable_amount=0.4000""",
        )
        self.assertEqual(
            intra_transaction.to_string(2, repr_format=False, extra_data=["foobar", "qwerty"]),
            """    IntraTransaction:
      id=19
      timestamp=2021-01-02 08:42:43.882000 +0000
      asset=B1
      from_exchange=Coinbase Pro
      from_holder=Bob
      to_exchange=BlockFi
      to_holder=Alice
      transaction_type=TransactionType.MOVE
      spot_price=1000.0000
      crypto_sent=2.00020000
      crypto_received=1.99980000
      crypto_fee=0.00040000
      fiat_fee=0.4000
      is_taxable=True
      fiat_taxable_amount=0.4000
      foobar
      qwerty""",
        )
        self.assertEqual(
            intra_transaction.to_string(2, repr_format=True, extra_data=["foobar", "qwerty"]),
            (
                "    IntraTransaction("
                "id='19', "
                "timestamp='2021-01-02 08:42:43.882000 +0000', "
                "asset='B1', "
                "from_exchange='Coinbase Pro', "
                "from_holder='Bob', "
                "to_exchange='BlockFi', "
                "to_holder='Alice', "
                "transaction_type=<TransactionType.MOVE: 'move'>, "
                "spot_price=1000.0000, "
                "crypto_sent=2.00020000, "
                "crypto_received=1.99980000, "
                "crypto_fee=0.00040000, "
                "fiat_fee=0.4000, "
                "is_taxable=True, "
                "fiat_taxable_amount=0.4000, "
                "foobar, "
                "qwerty)"
            ),
        )

    def test_non_taxable_intra_transaction(self) -> None:
        intra_transaction: IntraTransaction = IntraTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B2",
            "Coinbase",
            "Bob",
            "Coinbase",
            "Alice",
            RP2Decimal("100.0"),
            RP2Decimal("30"),
            RP2Decimal("30"),
            unique_id=19,
        )
        self.assertFalse(intra_transaction.is_taxable())
        self.assertEqual(RP2Decimal("0"), intra_transaction.fiat_taxable_amount)
        self.assertEqual("B2", intra_transaction.asset)
        self.assertEqual(TransactionType.MOVE, intra_transaction.transaction_type)
        self.assertEqual(RP2Decimal("0"), intra_transaction.crypto_fee)
        self.assertEqual(RP2Decimal("0"), intra_transaction.fiat_fee)
        self.assertEqual(RP2Decimal("-0"), intra_transaction.crypto_balance_change)
        self.assertEqual(RP2Decimal("0.0"), intra_transaction.fiat_balance_change)

        self.assertEqual(
            str(intra_transaction),
            """IntraTransaction:
  id=19
  timestamp=2021-01-02 08:42:43.882000 +0000
  asset=B2
  from_exchange=Coinbase
  from_holder=Bob
  to_exchange=Coinbase
  to_holder=Alice
  transaction_type=TransactionType.MOVE
  spot_price=100.0000
  crypto_sent=30.00000000
  crypto_received=30.00000000
  crypto_fee=0.00000000
  fiat_fee=0.0000
  is_taxable=False
  fiat_taxable_amount=0.0000""",
        )

    def test_intra_transaction_equality_and_hashing(self) -> None:
        intra_transaction: IntraTransaction = IntraTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B2",
            "Coinbase",
            "Bob",
            "Coinbase",
            "Alice",
            RP2Decimal("100.0"),
            RP2Decimal("30"),
            RP2Decimal("30"),
            unique_id=19,
        )
        intra_transaction2: IntraTransaction = IntraTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B2",
            "Coinbase",
            "Bob",
            "Coinbase",
            "Alice",
            RP2Decimal("100.0"),
            RP2Decimal("30"),
            RP2Decimal("30"),
            unique_id=19,
        )
        intra_transaction3: IntraTransaction = IntraTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B2",
            "Coinbase",
            "Bob",
            "Coinbase",
            "Alice",
            RP2Decimal("100.0"),
            RP2Decimal("30"),
            RP2Decimal("30"),
            unique_id=11,
        )
        self.assertEqual(intra_transaction, intra_transaction)
        self.assertEqual(intra_transaction, intra_transaction2)
        self.assertNotEqual(intra_transaction, intra_transaction3)
        self.assertEqual(hash(intra_transaction), hash(intra_transaction))
        self.assertEqual(hash(intra_transaction), hash(intra_transaction2))
        # These hashes would only be equal in case of hash collision (possible but very unlikey)
        self.assertNotEqual(hash(intra_transaction), hash(intra_transaction3))

    def test_bad_to_string(self) -> None:
        intra_transaction: IntraTransaction = IntraTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "Coinbase Pro",
            "Bob",
            "BlockFi",
            "Alice",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            RP2Decimal("1.9998"),
            unique_id=19,
        )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'indent' has non-integer value"):
            intra_transaction.to_string(None, repr_format=False, extra_data=["foobar", "qwerty"])  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'indent' has non-positive value.*"):
            intra_transaction.to_string(-1, repr_format=False, extra_data=["foobar", "qwerty"])
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'repr_format' has non-bool value .*"):
            intra_transaction.to_string(1, repr_format="False", extra_data=["foobar", "qwerty"])  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'extra_data' is not of type List"):
            intra_transaction.to_string(1, repr_format=False, extra_data="foobar")  # type: ignore

    def test_bad_intra_transaction(self) -> None:
        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string:.*"):
            IntraTransaction.type_check(None, None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_instance' is not of type IntraTransaction:.*"):
            IntraTransaction.type_check("my_instance", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_instance' is not of type IntraTransaction: InTransaction"):
            IntraTransaction.type_check(
                "my_instance",
                InTransaction(
                    self._configuration,
                    "2021-01-12T11:51:38Z",
                    "B1",
                    "BlockFi",
                    "Bob",
                    "BUY",
                    RP2Decimal("10000"),
                    RP2Decimal("1"),
                    RP2Decimal("0"),
                    unique_id=45,
                ),
            )

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            IntraTransaction(
                None,  # type: ignore
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            IntraTransaction(
                "config",  # type: ignore
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'unique_id' has non-positive value .*"):
            # Bad unique_id
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=-19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'unique_id' has non-integer value .*"):
            # Bad unique_id
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=(1, 2, 3),  # type: ignore
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'timestamp' has non-string value .*"):
            # Bad timestamp
            IntraTransaction(
                self._configuration,
                None,  # type: ignore
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'timestamp' value has no timezone info: .*"):
            # Bad timestamp
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'asset' value is not known: .*"):
            # Bad asset
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "hhh",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            # Bad asset
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                None,  # type: ignore
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'from_exchange' value is not known: .*"):
            # Bad from exchange
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "COinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'from_exchange' has non-string value .*"):
            # Bad from exchange
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                (1, 2, 3),  # type: ignore
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'from_holder' value is not known: .*"):
            # Bad from holder
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'from_holder' has non-string value .*"):
            # Bad from holder
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                None,  # type: ignore
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'to_exchange' value is not known: .*"):
            # Bad to exchange
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'to_exchange' has non-string value .*"):
            # Bad to exchange
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                None,  # type: ignore
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'to_holder' value is not known: .*"):
            # Bad to holder
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "A lice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'to_holder' has non-string value .*"):
            # Bad to holder
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                1111,  # type: ignore
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'spot_price' has non-positive value .*"):
            # Bad spot price
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("-1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'spot_price' has non-RP2Decimal value .*"):
            # Bad spot price
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                "1000",  # type: ignore
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_sent' has zero value"):
            # Bad crypto sent
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("0"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_sent' has non-positive value .*"):
            # Bad crypto sent
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("-2"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'crypto_sent' has non-RP2Decimal value .*"):
            # Bad crypto sent
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                None,  # type: ignore
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'crypto_sent' has non-RP2Decimal value .*"):
            # Bad crypto sent
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                "-2.0002",  # type: ignore
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_received' has non-positive value .*"):
            # Bad crypto received
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("-2"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'crypto_received' has non-RP2Decimal value .*"):
            # Bad crypto received
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                None,  # type: ignore
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'crypto_received' has non-RP2Decimal value .*"):
            # Bad crypto received
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                (1, 2, 3),  # type: ignore
                unique_id=19,
            )

        with self.assertRaisesRegex(RP2ValueError, "IntraTransaction .*, id.*from/to exchanges/holders are the same: sending to self"):
            # From/to exchanges/holders are the same: sending to self
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "BlockFi",
                "Bob",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_sent' has non-positive value .*"):
            # Crypto sent < 0
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("-2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_sent' has zero value"):
            # Crypto sent == 0
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("0"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_received' has non-positive value .*"):
            # Crypto received < 0
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("-1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'notes' has non-string value .*"):
            # Bad notes
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
                notes=1111,  # type: ignore
            )
        with self.assertRaisesRegex(RP2ValueError, ".*IntraTransaction .*, id.*: from/to exchanges/holders are the same: sending to self"):
            # Sender and receiver are the same
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Alice",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                RP2Decimal("1.9998"),
                unique_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, ".*IntraTransaction .*, id.*crypto sent < crypto received"):
            # Crypto sent < crypto received
            IntraTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "Coinbase Pro",
                "Bob",
                "BlockFi",
                "Alice",
                RP2Decimal("1000.0"),
                RP2Decimal("1.0002"),
                RP2Decimal("2.9998"),
                unique_id=19,
            )


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date
import os
import unittest
from enum import Enum
from pathlib import Path
from subprocess import run
from typing import List

from ods_diff import ods_diff
from rp2.configuration import MAX_DATE, MIN_DATE

ROOT_PATH: Path = Path(os.path.dirname(__file__)).parent.absolute()

CONFIG_PATH: Path = ROOT_PATH / Path("config")
INPUT_PATH: Path = ROOT_PATH / Path("input")
GOLDEN_PATH: Path = INPUT_PATH / Path("golden")


class OutputPlugins(Enum):
    RP2_FULL_REPORT = "rp2_full_report"
    TAX_REPORT_US = "tax_report_us"


class AbstractTestODSOutputDiff(unittest.TestCase):

    METHODS: List[str] = ["fifo", "lifo"]

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    @staticmethod
    def __get_time_interval(from_date: date = MIN_DATE, to_date: date = MAX_DATE) -> str:
        time_interval: str = ""
        if from_date > MIN_DATE and to_date < MAX_DATE:
            time_interval = f"{from_date}_{to_date}_"
        elif from_date > MIN_DATE and to_date >= MAX_DATE:
            time_interval = f"{from_date}_infinity_"
        elif from_date <= MIN_DATE and to_date < MAX_DATE:
            time_interval = f"0_{to_date}_"
        return time_interval

    @classmethod
    def _generate(
        cls,
        output_dir: Path,
        test_name: str,
        config: str,
        method: str,
        input_path: Path = INPUT_PATH,
        from_date: date = MIN_DATE,
        to_date: date = MAX_DATE,
    ) -> None:
        config = test_name if config is None else config
        time_interval: str = cls.__get_time_interval(from_date, to_date)

        arguments: List[str] = [
            "rp2_us",
            "-m",
            method,
            "-o",
            str(output_dir),
            "-p",
            f"{test_name}_{time_interval}",
        ]
        if from_date:
            arguments.extend(["-f", str(from_date)])
        if to_date:
            arguments.extend(["-t", str(to_date)])
        arguments.extend(
            [
                str(CONFIG_PATH / Path(f"{config}.config")),
                str(input_path / Path(f"{test_name}.ods")),
            ]
        )

        run(arguments, check=True)

    def _compare(
        self, output_dir: Path, test_name: str, method: str, output_plugin: OutputPlugins, from_date: date = MIN_DATE, to_date: date = MAX_DATE
    ) -> None:
        time_interval: str = self.__get_time_interval(from_date, to_date)
        diff: str
        output_file_name: Path = Path(f"{test_name}_{time_interval}{method}_{output_plugin.value}.ods")
        full_output_file_name: Path = output_dir / output_file_name
        full_golden_file_name: Path = GOLDEN_PATH / output_file_name
        diff = ods_diff(full_golden_file_name, full_output_file_name, generate_ascii_representation=True)
        self.assertFalse(diff, msg=diff)



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal, FloatOperation, getcontext

from rp2.rp2_error import RP2TypeError

CRYPTO_DECIMALS: int = 13
CRYPTO_DECIMAL_MASK: Decimal = Decimal("1." + "0" * int(CRYPTO_DECIMALS))

FIAT_DECIMALS: int = 2
FIAT_DECIMAL_MASK: Decimal = Decimal("1." + "0" * int(FIAT_DECIMALS))


class RP2Decimal(Decimal):

    # RP2Decimal initialization code. In Python there is no static constructor: the closest alternative is to add static initialization code
    # directly inside the class. Use arbitrarily high precision (quintillion + CRYPTO_DECIMALS digits)
    getcontext().prec = CRYPTO_DECIMALS + 18
    getcontext().traps[FloatOperation] = True

    @classmethod
    def is_equal_within_precision(cls, first: "RP2Decimal", second: "RP2Decimal", precision_mask: Decimal) -> bool:
        return (first - second).quantize(precision_mask) == ZERO

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return (self - other).quantize(CRYPTO_DECIMAL_MASK).__eq__(ZERO)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return (self - other).quantize(CRYPTO_DECIMAL_MASK).__ge__(ZERO)

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return (self - other).quantize(CRYPTO_DECIMAL_MASK).__gt__(ZERO)

    def __le__(self, other: object) -> bool:
        return not self.__gt__(other)

    def __lt__(self, other: object) -> bool:
        return not self.__ge__(other)

    def __add__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__add__(self, other))

    def __sub__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__sub__(self, other))

    def __mul__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__mul__(self, other))

    def __truediv__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__truediv__(self, other))

    def __floordiv__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__floordiv__(self, other))

    def __pow__(self, other: object, modulo: object = None) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        if modulo is not None and not isinstance(modulo, Decimal):
            raise RP2TypeError(f"Modulo has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__pow__(self, other, modulo))

    def __mod__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__mod__(self, other))

    # Reflected operand overrides
    def __radd__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__radd__(self, other))

    def __rsub__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__rsub__(self, other))

    def __rmul__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__rmul__(self, other))

    def __rtruediv__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__rtruediv__(self, other))

    def __rfloordiv__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__rfloordiv__(self, other))

    def __rmod__(self, other: object) -> "RP2Decimal":
        if not isinstance(other, Decimal):
            raise RP2TypeError(f"Operand has non-Decimal value {repr(other)}")
        return RP2Decimal(Decimal.__rmod__(self, other))


ZERO: RP2Decimal = RP2Decimal("0")



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from rp2.abstract_country import AbstractCountry
from rp2.rp2_main import rp2_main


# US-specific class
class US(AbstractCountry):
    def __init__(self) -> None:
        super().__init__("us", "usd")

    # Measured in days
    def long_term_capital_gain_period(self) -> int:
        return 365


# US-specific entry point
def rp2_entry() -> None:
    rp2_main(US())



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import unittest
from typing import List, Optional

from rp2.avl_tree import AVLTree


class TestRP2Decimal(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_insert(self) -> None:
        values: List[int]
        expected_representation: str
        for values, expected_representation in [
            ([10], "AVLTree(root=AVLNode(key=10, value=10, height=1, left=None, right=None))"),
            ([10, 20], "AVLTree(root=AVLNode(key=10, value=10, height=2, left=None, right=AVLNode(key=20, value=20, height=1, left=None, right=None)))"),
            (
                [10, 20, 30],
                (
                    "AVLTree(root=AVLNode(key=20, value=20, height=2, left=AVLNode(key=10, value=10, height=1, left=None, right=None), right=AVLNode(key=30, "
                    "value=30, height=1, left=None, right=None)))"
                ),
            ),
            (
                [10, 30, 20],
                (
                    "AVLTree(root=AVLNode(key=20, value=20, height=2, left=AVLNode(key=10, value=10, height=1, left=None, right=None), right=AVLNode(key=30, "
                    "value=30, height=1, left=None, right=None)))"
                ),
            ),
            (
                [20, 10, 30],
                (
                    "AVLTree(root=AVLNode(key=20, value=20, height=2, left=AVLNode(key=10, value=10, height=1, left=None, right=None), right=AVLNode(key=30, "
                    "value=30, height=1, left=None, right=None)))"
                ),
            ),
            (
                [20, 30, 10],
                (
                    "AVLTree(root=AVLNode(key=20, value=20, height=2, left=AVLNode(key=10, value=10, height=1, left=None, right=None), right=AVLNode(key=30, "
                    "value=30, height=1, left=None, right=None)))"
                ),
            ),
            (
                [30, 10, 20],
                (
                    "AVLTree(root=AVLNode(key=20, value=20, height=2, left=AVLNode(key=10, value=10, height=1, left=None, right=None), right=AVLNode(key=30, "
                    "value=30, height=1, left=None, right=None)))"
                ),
            ),
            (
                [30, 20, 10],
                (
                    "AVLTree(root=AVLNode(key=20, value=20, height=2, left=AVLNode(key=10, value=10, height=1, left=None, right=None), right=AVLNode(key=30, "
                    "value=30, height=1, left=None, right=None)))"
                ),
            ),
            (
                [10, 20, 30, 40],
                (
                    "AVLTree(root=AVLNode(key=20, value=20, height=3, left=AVLNode(key=10, value=10, height=1, left=None, right=None), right=AVLNode(key=30, "
                    "value=30, height=2, left=None, right=AVLNode(key=40, value=40, height=1, left=None, right=None))))"
                ),
            ),
            (
                [40, 30, 20, 10],
                (
                    "AVLTree(root=AVLNode(key=30, value=30, height=3, left=AVLNode(key=20, value=20, height=2, left=AVLNode(key=10, value=10, height=1, "
                    "left=None, right=None), right=None), right=AVLNode(key=40, value=40, height=1, left=None, right=None)))"
                ),
            ),
            (
                [10, 20, 30, 40, 50, 60, 70],
                (
                    "AVLTree(root=AVLNode(key=40, value=40, height=3, left=AVLNode(key=20, value=20, height=2, left=AVLNode(key=10, value=10, height=1, "
                    "left=None, right=None), right=AVLNode(key=30, value=30, height=1, left=None, right=None)), right=AVLNode(key=60, value=60, height=2, "
                    "left=AVLNode(key=50, value=50, height=1, left=None, right=None), right=AVLNode(key=70, value=70, height=1, left=None, right=None))))"
                ),
            ),
            (
                [48, 13, 92, 99, 2, 12, 6, 57, 22],
                (
                    "AVLTree(root=AVLNode(key=48, value=48, height=4, left=AVLNode(key=12, value=12, height=3, left=AVLNode(key=2, value=2, height=2, "
                    "left=None, right=AVLNode(key=6, value=6, height=1, left=None, right=None)), right=AVLNode(key=13, value=13, height=2, left=None, "
                    "right=AVLNode(key=22, value=22, height=1, left=None, right=None))), right=AVLNode(key=92, value=92, height=2, left=AVLNode(key=57, "
                    "value=57, height=1, left=None, right=None), right=AVLNode(key=99, value=99, height=1, left=None, right=None))))"
                ),
            ),
        ]:
            self._populate_with_numbers(values, expected_representation)

    def _populate_with_numbers(self, values: List[int], expected_representation: Optional[str] = None) -> AVLTree[int, int]:
        tree: AVLTree[int, int] = AVLTree()
        value: int
        for value in values:
            tree.insert_node(value, value)
        if expected_representation:
            self.assertEqual(repr(tree), expected_representation, f"{repr(tree)} != {expected_representation}")
        return tree

    def test_avl_tree_find(self) -> None:
        tree: AVLTree[int, int] = AVLTree()
        values: List[int] = [48, 13, 92, 99, 2, 12, 6, 57, 22]
        for value in values:
            tree.insert_node(value, value)

        for threshold, expected_value in [
            (0, None),
            (1, None),
            (2, 2),
            (9, 6),
            (22, 22),
            (72, 57),
            (98, 92),
            (100, 99),
            (1000, 99),
        ]:
            self.assertEqual(tree.find_max_value_less_than(threshold), expected_value, f"result != {expected_value}")


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger("abc").setLevel(logging.DEBUG)
    unittest.main()
